@@ -1,5 +1,5 @@
 /* chmod -- change permission modes of files
-   Copyright (C) 89, 90, 91, 95, 96, 97, 1998 Free Software Foundation, Inc.
+   Copyright (C) 89, 90, 91, 1995-2000 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -22,12 +22,17 @@
 #include <getopt.h>
 #include <sys/types.h>
 
-#include "modechange.h"
 #include "system.h"
-#include "closeout.h"
 #include "error.h"
-#include "savedir.h"
 #include "filemode.h"
+#include "modechange.h"
+#include "quote.h"
+#include "savedir.h"
+
+/* The official name of this program (e.g., no `g' prefix).  */
+#define PROGRAM_NAME "chmod"
+
+#define AUTHORS "David MacKenzie"
 
 enum Change_status
 {
@@ -70,11 +75,12 @@ static enum Verbosity verbosity = V_off;
    of this file.  This file must exist.  */
 static char *reference_file;
 
-/* If nonzero, display usage information and exit.  */
-static int show_help;
-
-/* If nonzero, print the version on standard output and exit.  */
-static int show_version;
+/* For long options that have no equivalent short option, use a
+   non-character as a pseudo short option, starting with CHAR_MAX + 1.  */
+enum
+{
+  REFERENCE_FILE_OPTION = CHAR_MAX + 1
+};
 
 static struct option const long_options[] =
 {
@@ -82,10 +88,10 @@ static struct option const long_options[] =
   {"changes", no_argument, 0, 'c'},
   {"silent", no_argument, 0, 'f'},
   {"quiet", no_argument, 0, 'f'},
-  {"reference", required_argument, 0, CHAR_MAX + 1},
+  {"reference", required_argument, 0, REFERENCE_FILE_OPTION},
   {"verbose", no_argument, 0, 'v'},
-  {"help", no_argument, &show_help, 1},
-  {"version", no_argument, &show_version, 1},
+  {GETOPT_HELP_OPTION_DECL},
+  {GETOPT_VERSION_OPTION_DECL},
   {0, 0, 0, 0}
 };
 
@@ -93,7 +99,7 @@ static struct option const long_options[] =
    CHANGED describes what (if anything) has happened. */
 
 static void
-describe_change (const char *file, short unsigned int mode,
+describe_change (const char *file, mode_t mode,
 		 enum Change_status changed)
 {
   char perms[11];		/* "-rwxrwxrwx" ls-style modes. */
@@ -104,18 +110,19 @@ describe_change (const char *file, short unsigned int mode,
   switch (changed)
     {
     case CH_SUCCEEDED:
-      fmt = _("mode of %s changed to %04o (%s)\n");
+      fmt = _("mode of %s changed to %04lo (%s)\n");
       break;
     case CH_FAILED:
-      fmt = _("failed to change mode of %s to %04o (%s)\n");
+      fmt = _("failed to change mode of %s to %04lo (%s)\n");
       break;
     case CH_NO_CHANGE_REQUESTED:
-      fmt = _("mode of %s retained as %04o (%s)\n");
+      fmt = _("mode of %s retained as %04lo (%s)\n");
       break;
     default:
       abort ();
     }
-  printf (fmt, file, mode & 07777, &perms[1]);
+  printf (fmt, quote (file),
+	  (unsigned long) (mode & CHMOD_MODE_BITS), &perms[1]);
 }
 
 /* Change the mode of FILE according to the list of operations CHANGES.
@@ -128,13 +135,13 @@ change_file_mode (const char *file, const struct mode_change *changes,
 		  const int deref_symlink)
 {
   struct stat file_stats;
-  unsigned short newmode;
+  mode_t newmode;
   int errors = 0;
 
   if (lstat (file, &file_stats))
     {
       if (force_silent == 0)
-	error (0, errno, "%s", file);
+	error (0, errno, _("getting attributes of %s"), quote (file));
       return 1;
     }
 #ifdef S_ISLNK
@@ -146,7 +153,7 @@ change_file_mode (const char *file, const struct mode_change *changes,
 	if (stat (file, &file_stats))
 	  {
 	    if (force_silent == 0)
-	      error (0, errno, "%s", file);
+	      error (0, errno, _("getting attributes of %s"), quote (file));
 	    return 1;
 	  }
     }
@@ -154,9 +161,10 @@ change_file_mode (const char *file, const struct mode_change *changes,
 
   newmode = mode_adjust (file_stats.st_mode, changes);
 
-  if (newmode != (file_stats.st_mode & 07777))
+  if (newmode != (file_stats.st_mode & CHMOD_MODE_BITS))
     {
-      int fail = chmod (file, (int) newmode);
+      int fail = chmod (file, newmode);
+      int saved_errno = errno;
 
       if (verbosity == V_high || (verbosity == V_changes_only && !fail))
 	describe_change (file, newmode, (fail ? CH_FAILED : CH_SUCCEEDED));
@@ -164,7 +172,8 @@ change_file_mode (const char *file, const struct mode_change *changes,
       if (fail)
 	{
 	  if (force_silent == 0)
-	    error (0, errno, "%s", file);
+	    error (0, saved_errno, _("changing permissions of %s"),
+		   quote (file));
 	  errors = 1;
 	}
     }
@@ -192,18 +201,12 @@ change_dir_mode (const char *dir, const struct mode_change *changes,
   unsigned pathlength;		/* Bytes allocated for `path'. */
   int errors = 0;
 
-  errno = 0;
-  name_space = savedir (dir, (unsigned int) statp->st_size);
+  name_space = savedir (dir, statp->st_size);
   if (name_space == NULL)
     {
-      if (errno)
-	{
-	  if (force_silent == 0)
-	    error (0, errno, "%s", dir);
-	  return 1;
-	}
-      else
-	error (1, 0, _("virtual memory exhausted"));
+      if (force_silent == 0)
+	error (0, errno, "%s", quote (dir));
+      return 1;
     }
 
   dirlength = strlen (dir) + 1;	/* + 1 is for the trailing '/'. */
@@ -229,7 +232,7 @@ change_dir_mode (const char *dir, const struct mode_change *changes,
   return errors;
 }
 
-static void
+void
 usage (int status)
 {
   if (status != 0)
@@ -239,11 +242,12 @@ usage (int status)
     {
       printf (_("\
 Usage: %s [OPTION]... MODE[,MODE]... FILE...\n\
-  or:  %s [OPTION]... OCTAL_MODE FILE...\n\
+  or:  %s [OPTION]... OCTAL-MODE FILE...\n\
   or:  %s [OPTION]... --reference=RFILE FILE...\n\
 "),
 	      program_name, program_name, program_name);
       printf (_("\
+Change the mode of each FILE to MODE.\n\
 \n\
   -c, --changes           like verbose but report only when a change is made\n\
   -f, --silent, --quiet   suppress most error messages\n\
@@ -257,7 +261,6 @@ Each MODE is one or more of the letters ugoa, one of the symbols +-= and\n\
 one or more of the letters rwxXstugo.\n\
 "));
       puts (_("\nReport bugs to <bug-fileutils@gnu.org>."));
-      close_stdout ();
     }
   exit (status);
 }
@@ -279,6 +282,8 @@ main (int argc, char **argv)
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
+  atexit (close_stdout);
+
   recurse = force_silent = 0;
 
   while (1)
@@ -311,7 +316,7 @@ main (int argc, char **argv)
 	    error (1, 0, _("invalid mode"));
 	  modeind = thisind;
 	  break;
-	case CHAR_MAX + 1:
+	case REFERENCE_FILE_OPTION:
 	  reference_file = optarg;
 	  break;
 	case 'R':
@@ -326,21 +331,13 @@ main (int argc, char **argv)
 	case 'v':
 	  verbosity = V_high;
 	  break;
+	case_GETOPT_HELP_CHAR;
+	case_GETOPT_VERSION_CHAR (PROGRAM_NAME, AUTHORS);
 	default:
 	  usage (1);
 	}
     }
 
-  if (show_version)
-    {
-      printf ("chmod (%s) %s\n", GNU_PACKAGE, VERSION);
-      close_stdout ();
-      exit (0);
-    }
-
-  if (show_help)
-    usage (0);
-
   if (modeind == 0 && reference_file == NULL)
     modeind = optind++;
 
@@ -356,9 +353,9 @@ main (int argc, char **argv)
   if (changes == MODE_INVALID)
     error (1, 0, _("invalid mode"));
   else if (changes == MODE_MEMORY_EXHAUSTED)
-    error (1, 0, _("virtual memory exhausted"));
+    xalloc_die ();
   else if (changes == MODE_BAD_REFERENCE)
-    error (1, errno, "%s", reference_file);
+    error (1, errno, _("getting attributes of %s"), quote (reference_file));
 
   for (; optind < argc; ++optind)
     {
@@ -366,7 +363,5 @@ main (int argc, char **argv)
       errors |= change_file_mode (argv[optind], changes, 1);
     }
 
-  if (verbosity != V_off)
-    close_stdout ();
   exit (errors);
 }
