@@ -1,5 +1,5 @@
 /* cp.c  -- file copying (main routines)
-   Copyright (C) 89, 90, 91, 95, 96, 97, 1998 Free Software Foundation.
+   Copyright (C) 89, 90, 91, 1995-2000 Free Software Foundation.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -28,13 +28,29 @@
 #include <getopt.h>
 
 #include "system.h"
-#include "backupfile.h"
 #include "argmatch.h"
-#include "path-concat.h"
-#include "closeout.h"
-#include "cp-hash.h"
+#include "backupfile.h"
 #include "copy.h"
+#include "cp-hash.h"
 #include "error.h"
+#include "dirname.h"
+#include "path-concat.h"
+#include "quote.h"
+
+#define ASSIGN_BASENAME_STRDUPA(Dest, File_name)	\
+  do							\
+    {							\
+      char *tmp_abns_;					\
+      ASSIGN_STRDUPA (tmp_abns_, (File_name));		\
+      strip_trailing_slashes (tmp_abns_);		\
+      Dest = base_name (tmp_abns_);			\
+    }							\
+  while (0)
+
+/* The official name of this program (e.g., no `g' prefix).  */
+#define PROGRAM_NAME "cp"
+
+#define AUTHORS "Torbjorn Granlund, David MacKenzie, and Jim Meyering"
 
 #ifndef _POSIX_VERSION
 uid_t geteuid ();
@@ -50,14 +66,18 @@ struct dir_attr
   struct dir_attr *next;
 };
 
-int stat ();
-int lstat ();
+/* For long options that have no equivalent short option, use a
+   non-character as a pseudo short option, starting with CHAR_MAX + 1.  */
+enum
+{
+  TARGET_DIRECTORY_OPTION = CHAR_MAX + 1,
+  SPARSE_OPTION,
+  STRIP_TRAILING_SLASHES_OPTION,
+  PARENTS_OPTION,
+  UNLINK_DEST_BEFORE_OPENING
+};
 
-char *base_name ();
-char *dirname ();
-enum backup_type get_version ();
 void strip_trailing_slashes ();
-char *xstrdup ();
 
 /* Initial number of entries in each hash table entry's table of inodes.  */
 #define INITIAL_HASH_MODULE 100
@@ -72,6 +92,9 @@ char *program_name;
    as its destination instead of the usual "e_dir/e_file." */
 static int flag_path = 0;
 
+/* Remove any trailing slashes from each SOURCE argument.  */
+static int remove_trailing_slashes;
+
 static char const *const sparse_type_string[] =
 {
   "never", "auto", "always", 0
@@ -85,37 +108,35 @@ static enum Sparse_type const sparse_type[] =
 /* The error code to return to the system. */
 static int exit_status = 0;
 
-/* If nonzero, display usage information and exit.  */
-static int show_help;
-
-/* If nonzero, print the version on standard output and exit.  */
-static int show_version;
-
 static struct option const long_opts[] =
 {
   {"archive", no_argument, NULL, 'a'},
-  {"backup", no_argument, NULL, 'b'},
+  {"backup", optional_argument, NULL, 'b'},
+  {"dereference", no_argument, NULL, 'L'},
   {"force", no_argument, NULL, 'f'},
-  {"sparse", required_argument, NULL, CHAR_MAX + 1},
+  {"sparse", required_argument, NULL, SPARSE_OPTION},
   {"interactive", no_argument, NULL, 'i'},
   {"link", no_argument, NULL, 'l'},
   {"no-dereference", no_argument, NULL, 'd'},
   {"one-file-system", no_argument, NULL, 'x'},
-  {"parents", no_argument, NULL, 'P'},
-  {"path", no_argument, NULL, 'P'},
+  {"parents", no_argument, NULL, PARENTS_OPTION},
+  {"path", no_argument, NULL, PARENTS_OPTION},   /* Deprecated.  */
   {"preserve", no_argument, NULL, 'p'},
   {"recursive", no_argument, NULL, 'R'},
+  {"remove-destination", no_argument, NULL, UNLINK_DEST_BEFORE_OPENING},
+  {"strip-trailing-slash", no_argument, NULL, STRIP_TRAILING_SLASHES_OPTION},
   {"suffix", required_argument, NULL, 'S'},
   {"symbolic-link", no_argument, NULL, 's'},
+  {"target-directory", required_argument, NULL, TARGET_DIRECTORY_OPTION},
   {"update", no_argument, NULL, 'u'},
   {"verbose", no_argument, NULL, 'v'},
-  {"version-control", required_argument, NULL, 'V'},
-  {"help", no_argument, &show_help, 1},
-  {"version", no_argument, &show_version, 1},
+  {"version-control", required_argument, NULL, 'V'}, /* Deprecated. FIXME. */
+  {GETOPT_HELP_OPTION_DECL},
+  {GETOPT_VERSION_OPTION_DECL},
   {NULL, 0, NULL, 0}
 };
 
-static void
+void
 usage (int status)
 {
   if (status != 0)
@@ -126,29 +147,42 @@ usage (int status)
       printf (_("\
 Usage: %s [OPTION]... SOURCE DEST\n\
   or:  %s [OPTION]... SOURCE... DIRECTORY\n\
+  or:  %s [OPTION]... --target-directory=DIRECTORY SOURCE...\n\
 "),
-	      program_name, program_name);
+	      program_name, program_name, program_name);
       printf (_("\
 Copy SOURCE to DEST, or multiple SOURCE(s) to DIRECTORY.\n\
 \n\
   -a, --archive                same as -dpR\n\
-  -b, --backup                 make backup before removal\n\
-  -d, --no-dereference         preserve links\n\
-  -f, --force                  remove existing destinations, never prompt\n\
+      --backup[=CONTROL]       make a backup of each existing destination file\n\
+  -b                           like --backup but does not accept an argument\n\
+  -d, --no-dereference         never follow symbolic links\n\
+  -f, --force                  if an existing destination file cannot be\n\
+                                   opened, remove it and try again\n\
   -i, --interactive            prompt before overwrite\n\
+  -H                           follow command-line symbolic links\n\
   -l, --link                   link files instead of copying\n\
+  -L, --dereference            always follow symbolic links\n\
   -p, --preserve               preserve file attributes if possible\n\
-  -P, --parents                append source path to DIRECTORY\n\
+      --parents                append source path to DIRECTORY\n\
+  -P                           same as `--parents' for now; soon to change to\n\
+                                 `--no-dereference' to conform to POSIX\n\
   -r                           copy recursively, non-directories as files\n\
+                                 WARNING: use -R instead when you might copy\n\
+                                 special files like FIFOs or /dev/zero\n\
+      --remove-destination     remove each existing destination file before\n\
+                                 attempting to open it (contrast with --force)\n\
       --sparse=WHEN            control creation of sparse files\n\
   -R, --recursive              copy directories recursively\n\
+      --strip-trailing-slashes remove any trailing slashes from each SOURCE\n\
+                                 argument\n\
   -s, --symbolic-link          make symbolic links instead of copying\n\
   -S, --suffix=SUFFIX          override the usual backup suffix\n\
+      --target-directory=DIRECTORY  move all SOURCE arguments into DIRECTORY\n\
   -u, --update                 copy only when the SOURCE file is newer\n\
                                  than the destination file or when the\n\
                                  destination file is missing\n\
   -v, --verbose                explain what is being done\n\
-  -V, --version-control=WORD   override the usual version control\n\
   -x, --one-file-system        stay on this file system\n\
       --help                   display this help and exit\n\
       --version                output version information and exit\n\
@@ -161,12 +195,14 @@ Use --sparse=never to inhibit creation of sparse files.\n\
 \n\
 "));
       printf (_("\
-The backup suffix is ~, unless set with SIMPLE_BACKUP_SUFFIX.  The\n\
-version control may be set with VERSION_CONTROL, values are:\n\
+The backup suffix is `~', unless set with --suffix or SIMPLE_BACKUP_SUFFIX.\n\
+The version control method may be selected via the --backup option or through\n\
+the VERSION_CONTROL environment variable.  Here are the values:\n\
 \n\
-  t, numbered     make numbered backups\n\
-  nil, existing   numbered if numbered backups exist, simple otherwise\n\
-  never, simple   always make simple backups\n\
+  none, off       never make backups (even if --backup is given)\n\
+  numbered, t     make numbered backups\n\
+  existing, nil   numbered if numbered backups exist, simple otherwise\n\
+  simple, never   always make simple backups\n\
 "));
       printf (_("\
 \n\
@@ -175,7 +211,6 @@ options are given and SOURCE and DEST are the same name for an existing,\n\
 regular file.\n\
 "));
       puts (_("\nReport bugs to <bug-fileutils@gnu.org>."));
-      close_stdout ();
     }
   exit (status);
 }
@@ -221,7 +256,8 @@ re_protect (const char *const_dst_path, int src_offset,
 
       if ((*(x->xstat)) (src_path, &src_sb))
 	{
-	  error (0, errno, "%s", src_path);
+	  error (0, errno, _("getting attributes of %s"),
+		 quote (src_path));
 	  return 1;
 	}
 
@@ -242,7 +278,7 @@ re_protect (const char *const_dst_path, int src_offset,
 
 	  if (utime (dst_path, &utb))
 	    {
-	      error (0, errno, _("preserving times for %s"), dst_path);
+	      error (0, errno, _("preserving times for %s"), quote (dst_path));
 	      return 1;
 	    }
 	}
@@ -255,7 +291,8 @@ re_protect (const char *const_dst_path, int src_offset,
 	  if (chown (dst_path, src_sb.st_uid, src_sb.st_gid)
 	      && ((errno != EPERM && errno != EINVAL) || myeuid == 0))
 	    {
-	      error (0, errno, _("preserving ownership for %s"), dst_path);
+	      error (0, errno, _("preserving ownership for %s"),
+		     quote (dst_path));
 	      return 1;
 	    }
 	}
@@ -264,7 +301,8 @@ re_protect (const char *const_dst_path, int src_offset,
 	{
 	  if (chmod (dst_path, src_sb.st_mode & x->umask_kill))
 	    {
-	      error (0, errno, _("preserving permissions for %s"), dst_path);
+	      error (0, errno, _("preserving permissions for %s"),
+		     quote (dst_path));
 	      return 1;
 	    }
 	}
@@ -308,7 +346,7 @@ make_path_private (const char *const_dirpath, int src_offset, int mode,
 
   src = dirpath + src_offset;
 
-  tmp_dst_dirname = dirname (dirpath);
+  tmp_dst_dirname = dir_name (dirpath);
   dst_dirname = (char *) alloca (strlen (tmp_dst_dirname) + 1);
   strcpy (dst_dirname, tmp_dst_dirname);
   free (tmp_dst_dirname);
@@ -346,7 +384,8 @@ make_path_private (const char *const_dirpath, int src_offset, int mode,
 	      new->is_new_dir = 1;
 	      if (mkdir (dirpath, mode))
 		{
-		  error (0, errno, _("cannot make directory `%s'"), dirpath);
+		  error (0, errno, _("cannot make directory %s"),
+			 quote (dirpath));
 		  return 1;
 		}
 	      else
@@ -357,7 +396,8 @@ make_path_private (const char *const_dirpath, int src_offset, int mode,
 	    }
 	  else if (!S_ISDIR (stats.st_mode))
 	    {
-	      error (0, 0, _("`%s' exists but is not a directory"), dirpath);
+	      error (0, 0, _("%s exists but is not a directory"),
+		     quote (dirpath));
 	      return 1;
 	    }
 	  else
@@ -378,7 +418,7 @@ make_path_private (const char *const_dirpath, int src_offset, int mode,
 
   else if (!S_ISDIR (stats.st_mode))
     {
-      error (0, 0, _("`%s' exists but is not a directory"), dst_dirname);
+      error (0, 0, _("%s exists but is not a directory"), quote (dst_dirname));
       return 1;
     }
   else
@@ -392,35 +432,43 @@ make_path_private (const char *const_dirpath, int src_offset, int mode,
    Return 0 if successful, 1 if any errors occur. */
 
 static int
-do_copy (int argc, char **argv, const struct cp_options *x)
+do_copy (int n_files, char **file, const char *target_directory,
+	 const struct cp_options *x)
 {
-  char *dest;
+  const char *dest;
   struct stat sb;
   int new_dst = 0;
   int ret = 0;
+  int dest_is_dir = 0;
 
-  if (optind >= argc)
+  if (n_files <= 0)
     {
       error (0, 0, _("missing file arguments"));
       usage (1);
     }
-  if (optind >= argc - 1)
+  if (n_files == 1 && !target_directory)
     {
       error (0, 0, _("missing destination file"));
       usage (1);
     }
 
-  dest = argv[argc - 1];
+  if (target_directory)
+    dest = target_directory;
+  else
+    {
+      dest = file[n_files - 1];
+      --n_files;
+    }
 
   if (lstat (dest, &sb))
     {
       if (errno != ENOENT)
 	{
-	  error (0, errno, "%s", dest);
+	  error (0, errno, _("accessing %s"), quote (dest));
 	  return 1;
 	}
-      else
-	new_dst = 1;
+
+      new_dst = 1;
     }
   else
     {
@@ -431,40 +479,59 @@ do_copy (int argc, char **argv, const struct cp_options *x)
 	 into symlinks to directories. */
       if (stat (dest, &sbx) == 0)
 	sb = sbx;
+
+      dest_is_dir = S_ISDIR (sb.st_mode);
     }
 
-  if (!new_dst && S_ISDIR (sb.st_mode))
+  if (!dest_is_dir)
+    {
+      if (target_directory)
+	{
+	  error (0, 0, _("specified target, %s is not a directory"),
+		 quote (dest));
+	  usage (1);
+	}
+
+      if (n_files > 1)
+	{
+	  error (0, 0,
+	 _("copying multiple files, but last argument %s is not a directory"),
+	     quote (dest));
+	  usage (1);
+	}
+    }
+
+  if (dest_is_dir)
     {
       /* cp file1...filen edir
 	 Copy the files `file1' through `filen'
 	 to the existing directory `edir'. */
+      int i;
 
-      for (;;)
+      for (i = 0; i < n_files; i++)
 	{
-	  char *arg;
-	  char *ap;
 	  char *dst_path;
-	  int parent_exists = 1; /* True if dirname (dst_path) exists. */
+	  int parent_exists = 1; /* True if dir_name (dst_path) exists. */
 	  struct dir_attr *attr_list;
 	  char *arg_in_concat = NULL;
+	  char *arg = file[i];
 
-	  arg = argv[optind];
-
-	  strip_trailing_slashes (arg);
+	  if (remove_trailing_slashes)
+	    strip_trailing_slashes (arg);
 
 	  if (flag_path)
 	    {
 	      /* Append all of `arg' to `dest'.  */
 	      dst_path = path_concat (dest, arg, &arg_in_concat);
 	      if (dst_path == NULL)
-		error (1, 0, _("virtual memory exhausted"));
+		xalloc_die ();
 
 	      /* For --parents, we have to make sure that the directory
-	         dirname (dst_path) exists.  We may have to create a few
+	         dir_name (dst_path) exists.  We may have to create a few
 	         leading directories. */
 	      parent_exists = !make_path_private (dst_path,
 						  arg_in_concat - dst_path,
-						  0700,
+						  S_IRWXU,
 						  (x->verbose
 						   ? "%s -> %s\n" : NULL),
 						  &attr_list, &new_dst,
@@ -472,13 +539,14 @@ do_copy (int argc, char **argv, const struct cp_options *x)
 	    }
   	  else
   	    {
+	      char *arg_base;
 	      /* Append the last component of `arg' to `dest'.  */
 
-	      ap = base_name (arg);
+	      ASSIGN_BASENAME_STRDUPA (arg_base, arg);
 	      /* For `cp -R source/.. dest', don't copy into `dest/..'. */
-	      dst_path = (STREQ (ap, "..")
+	      dst_path = (STREQ (arg_base, "..")
 			  ? xstrdup (dest)
-			  : path_concat (dest, ap, NULL));
+			  : path_concat (dest, arg_base, NULL));
 	    }
 
 	  if (!parent_exists)
@@ -488,8 +556,8 @@ do_copy (int argc, char **argv, const struct cp_options *x)
   	    }
 	  else
 	    {
-	      int unused;
-	      ret |= copy (arg, dst_path, new_dst, x, &unused, NULL);
+	      int copy_into_self;
+	      ret |= copy (arg, dst_path, new_dst, x, &copy_into_self, NULL);
 	      forget_all ();
 
 	      if (flag_path)
@@ -500,13 +568,10 @@ do_copy (int argc, char **argv, const struct cp_options *x)
 	    }
 
 	  free (dst_path);
-	  ++optind;
-	  if (optind == argc - 1)
-	    break;
 	}
       return ret;
     }
-  else if (argc - optind == 2)
+  else /* if (n_files == 1) */
     {
       char *new_dest;
       char *source;
@@ -516,11 +581,11 @@ do_copy (int argc, char **argv, const struct cp_options *x)
       if (flag_path)
 	{
 	  error (0, 0,
-	       _("when preserving paths, last argument must be a directory"));
+	       _("when preserving paths, the destination must be a directory"));
 	  usage (1);
 	}
 
-      source = argv[optind];
+      source = file[0];
 
       /* When the force and backup options have been specified and
 	 the source and destination are the same name for an existing
@@ -528,7 +593,7 @@ do_copy (int argc, char **argv, const struct cp_options *x)
 	 `cp --force --backup foo foo' to `cp --force foo fooSUFFIX'
 	 where SUFFIX is determined by any version control options used.  */
 
-      if (x->force
+      if (x->unlink_dest_after_failed_open
 	  && x->backup_type != none
 	  && STREQ (source, dest)
 	  && !new_dst && S_ISREG (sb.st_mode))
@@ -546,7 +611,7 @@ do_copy (int argc, char **argv, const struct cp_options *x)
 	  x = &x_tmp;
 
 	  if (new_dest == NULL)
-	    error (1, 0, _("virtual memory exhausted"));
+	    xalloc_die ();
 	}
 
       /* When the destination is specified with a trailing slash and the
@@ -561,43 +626,30 @@ do_copy (int argc, char **argv, const struct cp_options *x)
 	  && !S_ISDIR (source_stats.st_mode))
 	{
 	  char *source_base;
-	  char *tmp_source;
-
-	  tmp_source = (char *) alloca (strlen (source) + 1);
-	  strcpy (tmp_source, source);
-	  strip_trailing_slashes (tmp_source);
-	  source_base = base_name (tmp_source);
 
+	  ASSIGN_BASENAME_STRDUPA (source_base, source);
 	  new_dest = (char *) alloca (strlen (dest)
 				      + strlen (source_base) + 1);
 	  stpcpy (stpcpy (new_dest, dest), source_base);
 	}
       else
 	{
-	  new_dest = dest;
+	  new_dest = (char *) dest;
 	}
 
       return copy (source, new_dest, new_dst, x, &unused, NULL);
     }
-  else
-    {
-      error (0, 0,
-	     _("copying multiple files, but last argument (%s) \
-is not a directory"),
-	     dest);
-      usage (1);
-    }
 
   /* unreachable */
-  return 0;
 }
 
 static void
 cp_option_init (struct cp_options *x)
 {
   x->copy_as_regular = 1;
-  x->dereference = 1;
-  x->force = 0;
+  x->dereference = DEREF_UNDEFINED;
+  x->unlink_dest_before_opening = 0;
+  x->unlink_dest_after_failed_open = 0;
   x->failed_unlink_is_fatal = 1;
   x->hard_link = 0;
   x->interactive = 0;
@@ -620,7 +672,7 @@ cp_option_init (struct cp_options *x)
      when using chmod.  The creation mask is set to be liberal, so
      that created directories can be written, even if it would not
      have been allowed with the mask this process was started with.  */
-  x->umask_kill = 0777777 ^ umask (0);
+  x->umask_kill = ~ umask (0);
 
   x->update = 0;
   x->verbose = 0;
@@ -631,22 +683,26 @@ main (int argc, char **argv)
 {
   int c;
   int make_backups = 0;
-  char *version;
+  char *backup_suffix_string;
+  char *version_control_string = NULL;
   struct cp_options x;
+  char *target_directory = NULL;
+  int used_P_option = 0;
 
   program_name = argv[0];
   setlocale (LC_ALL, "");
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
+  atexit (close_stdout);
+
   cp_option_init (&x);
 
-  version = getenv ("SIMPLE_BACKUP_SUFFIX");
-  if (version)
-    simple_backup_suffix = version;
-  version = getenv ("VERSION_CONTROL");
+  /* FIXME: consider not calling getenv for SIMPLE_BACKUP_SUFFIX unless
+     we'll actually use backup_suffix_string.  */
+  backup_suffix_string = getenv ("SIMPLE_BACKUP_SUFFIX");
 
-  while ((c = getopt_long (argc, argv, "abdfilprsuvxPRS:V:", long_opts, NULL))
+  while ((c = getopt_long (argc, argv, "abdfHilLprsuvxPRS:V:", long_opts, NULL))
 	 != -1)
     {
       switch (c)
@@ -654,23 +710,13 @@ main (int argc, char **argv)
 	case 0:
 	  break;
 
-	case CHAR_MAX + 1:
-	  {
-	    int i;
-
-	    /* --sparse={never,auto,always}  */
-	    i = argmatch (optarg, sparse_type_string);
-	    if (i < 0)
-	      {
-		invalid_arg (_("sparse type"), optarg, i);
-		usage (1);
-	      }
-	    x.sparse_mode = sparse_type[i];
-	  }
+	case SPARSE_OPTION:
+	  x.sparse_mode = XARGMATCH ("--sparse", optarg,
+				     sparse_type_string, sparse_type);
 	  break;
 
 	case 'a':		/* Like -dpR. */
-	  x.dereference = 0;
+	  x.dereference = DEREF_NEVER;
 	  x.preserve_owner_and_group = 1;
 	  x.preserve_chmod_bits = 1;
 	  x.preserve_timestamps = 1;
@@ -679,21 +725,32 @@ main (int argc, char **argv)
 	  x.copy_as_regular = 0;
 	  break;
 
+	case 'V':  /* FIXME: this is deprecated.  Remove it in 2001.  */
+	  error (0, 0,
+		 _("warning: --version-control (-V) is obsolete;  support for\
+ it\nwill be removed in some future release.  Use --backup=%s instead."
+		   ), optarg);
+	  /* Fall through.  */
+
 	case 'b':
 	  make_backups = 1;
+	  if (optarg)
+	    version_control_string = optarg;
 	  break;
 
 	case 'd':
-	  x.dereference = 0;
+	  x.dereference = DEREF_NEVER;
 	  break;
 
 	case 'f':
-	  x.force = 1;
-	  x.interactive = 0;
+	  x.unlink_dest_after_failed_open = 1;
+	  break;
+
+	case 'H':
+	  x.dereference = DEREF_COMMAND_LINE_ARGUMENTS;
 	  break;
 
 	case 'i':
-	  x.force = 0;
 	  x.interactive = 1;
 	  break;
 
@@ -701,6 +758,10 @@ main (int argc, char **argv)
 	  x.hard_link = 1;
 	  break;
 
+	case 'L':
+	  x.dereference = DEREF_ALWAYS;
+	  break;
+
 	case 'p':
 	  x.preserve_owner_and_group = 1;
 	  x.preserve_chmod_bits = 1;
@@ -709,6 +770,9 @@ main (int argc, char **argv)
 	  break;
 
 	case 'P':
+	  used_P_option = 1;
+	  /* fall through */
+	case PARENTS_OPTION:
 	  flag_path = 1;
 	  break;
 
@@ -722,6 +786,14 @@ main (int argc, char **argv)
 	  x.copy_as_regular = 0;
 	  break;
 
+	case UNLINK_DEST_BEFORE_OPENING:
+	  x.unlink_dest_before_opening = 1;
+	  break;
+
+	case STRIP_TRAILING_SLASHES_OPTION:
+	  remove_trailing_slashes = 1;
+	  break;
+
 	case 's':
 #ifdef S_ISLNK
 	  x.symbolic_link = 1;
@@ -730,6 +802,10 @@ main (int argc, char **argv)
 #endif
 	  break;
 
+	case TARGET_DIRECTORY_OPTION:
+	  target_directory = optarg;
+	  break;
+
 	case 'u':
 	  x.update = 1;
 	  break;
@@ -743,54 +819,76 @@ main (int argc, char **argv)
 	  break;
 
 	case 'S':
-	  simple_backup_suffix = optarg;
+	  make_backups = 1;
+	  backup_suffix_string = optarg;
 	  break;
 
-	case 'V':
-	  version = optarg;
-	  break;
+	case_GETOPT_HELP_CHAR;
+
+	case_GETOPT_VERSION_CHAR (PROGRAM_NAME, AUTHORS);
 
 	default:
 	  usage (1);
 	}
     }
 
-  if (show_version)
-    {
-      printf ("cp (%s) %s\n", GNU_PACKAGE, VERSION);
-      close_stdout ();
-      exit (0);
-    }
-
-  if (show_help)
-    usage (0);
-
   if (x.hard_link && x.symbolic_link)
     {
       error (0, 0, _("cannot make both hard and symbolic links"));
       usage (1);
     }
 
-  x.backup_type = (make_backups ? get_version (version) : none);
+  if (used_P_option)
+    {
+      error (0, 0,
+	     _("\
+Warning: the meaning of `-P' will change in the future to conform to POSIX.\n\
+Use `--parents' for the old meaning, and `--no-dereference' for the new one."));
+    }
+
+  if (backup_suffix_string)
+    simple_backup_suffix = xstrdup (backup_suffix_string);
+
+  x.backup_type = (make_backups
+		   ? xget_version (_("backup type"),
+				   version_control_string)
+		   : none);
 
   if (x.preserve_chmod_bits == 1)
-    x.umask_kill = 0777777;
+    x.umask_kill = ~ (mode_t) 0;
+
+  if (x.dereference == DEREF_UNDEFINED)
+    {
+      if (x.recursive)
+	/* This is compatible with FreeBSD.  */
+	x.dereference = DEREF_NEVER;
+      else
+	x.dereference = DEREF_ALWAYS;
+    }
 
   /* The key difference between -d (--no-dereference) and not is the version
      of `stat' to call.  */
 
-  if (x.dereference)
-    x.xstat = stat;
-  else
+  if (x.dereference == DEREF_NEVER)
     x.xstat = lstat;
+  else
+    {
+      /* For DEREF_COMMAND_LINE_ARGUMENTS, x.xstat must be stat for
+	 each command line argument, but must later be `lstat' for
+	 any symlinks that are found via recursive traversal.  */
+      x.xstat = stat;
+    }
+
+  /* If --force (-f) was specified and we're in link-creation mode,
+     first remove any existing destination file.  */
+  if (x.unlink_dest_after_failed_open && (x.hard_link || x.symbolic_link))
+    x.unlink_dest_before_opening = 1;
 
   /* Allocate space for remembering copied and created files.  */
 
   hash_init (INITIAL_HASH_MODULE, INITIAL_ENTRY_TAB_SIZE);
 
-  exit_status |= do_copy (argc, argv, &x);
+  exit_status |= do_copy (argc - optind, argv + optind, target_directory, &x);
 
-  if (x.verbose)
-    close_stdout ();
   exit (exit_status);
 }
