@@ -1,5 +1,5 @@
 /* dd -- convert a file while copying it.
-   Copyright (C) 85, 90, 91, 95, 96, 97, 1998 Free Software Foundation, Inc.
+   Copyright (C) 85, 90, 91, 1995-2000 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -29,16 +29,29 @@
 #include <signal.h>
 #include <getopt.h>
 
-#include "human.h"
 #include "system.h"
 #include "closeout.h"
 #include "error.h"
+#include "getpagesize.h"
+#include "human.h"
+#include "long-options.h"
+#include "quote.h"
 #include "safe-read.h"
+#include "xstrtol.h"
+
+/* The official name of this program (e.g., no `g' prefix).  */
+#define PROGRAM_NAME "dd"
+
+#define AUTHORS "Paul Rubin, David MacKenzie, and Stuart Kemp"
 
 #ifndef SIGINFO
 # define SIGINFO SIGUSR1
 #endif
 
+#define ROUND_UP_OFFSET(X, M) ((M) - 1 - (((X) + (M) - 1) % (M)))
+#define PTR_ALIGN(Ptr, M) ((Ptr) \
+			   + ROUND_UP_OFFSET ((char *)(Ptr) - (char *)0, (M)))
+
 #define max(a, b) ((a) > (b) ? (a) : (b))
 #define output_char(c) \
   do { \
@@ -71,15 +84,9 @@ char *program_name;
 /* The name of the input file, or NULL for the standard input. */
 static char *input_file = NULL;
 
-/* The input file descriptor. */
-static int input_fd = 0;
-
 /* The name of the output file, or NULL for the standard output. */
 static char *output_file = NULL;
 
-/* The output file descriptor. */
-static int output_fd = 1;
-
 /* The number of bytes in which atomic reads are done. */
 static size_t input_blocksize = 0;
 
@@ -266,20 +273,7 @@ static unsigned char const ebcdic_to_ascii[] =
   070, 071, 0372, 0373, 0374, 0375, 0376, 0377
 };
 
-/* If nonzero, display usage information and exit.  */
-static int show_help;
-
-/* If nonzero, print the version on standard output and exit.  */
-static int show_version;
-
-static struct option const long_options[] =
-{
-  {"help", no_argument, &show_help, 1},
-  {"version", no_argument, &show_version, 1},
-  {0, 0, 0, 0}
-};
-
-static void
+void
 usage (int status)
 {
   if (status != 0)
@@ -322,7 +316,6 @@ Each KEYWORD may be:\n\
   sync      pad every input block with NULs to ibs-size\n\
 "));
       puts (_("\nReport bugs to <bug-fileutils@gnu.org>."));
-      close_stdout ();
     }
   exit (status);
 }
@@ -330,7 +323,7 @@ Each KEYWORD may be:\n\
 static void
 translate_charset (const unsigned char *new_trans)
 {
-  int i;
+  unsigned int i;
 
   for (i = 0; i < 256; i++)
     trans_table[i] = new_trans[trans_table[i]];
@@ -373,10 +366,10 @@ static void
 cleanup (void)
 {
   print_stats ();
-  if (close (input_fd) < 0)
-    error (1, errno, "%s", input_file);
-  if (close (output_fd) < 0)
-    error (1, errno, "%s", output_file);
+  if (close (STDIN_FILENO) < 0)
+    error (1, errno, _("closing input file %s"), quote (input_file));
+  if (close (STDOUT_FILENO) < 0)
+    error (1, errno, _("closing output file %s"), quote (output_file));
 }
 
 static inline void
@@ -389,22 +382,22 @@ quit (int code)
 static RETSIGTYPE
 interrupt_handler (int sig)
 {
-#ifdef SA_INTERRUPT
+#ifdef SA_NOCLDSTOP
   struct sigaction sigact;
 
   sigact.sa_handler = SIG_DFL;
   sigemptyset (&sigact.sa_mask);
   sigact.sa_flags = 0;
   sigaction (sig, &sigact, NULL);
-#else				/* !SA_INTERRUPT */
+#else
   signal (sig, SIG_DFL);
-#endif				/* SA_INTERRUPT */
+#endif
   cleanup ();
   kill (getpid (), sig);
 }
 
 static RETSIGTYPE
-siginfo_handler (int sig)
+siginfo_handler (int sig ATTRIBUTE_UNUSED)
 {
   print_stats ();
 }
@@ -414,7 +407,7 @@ siginfo_handler (int sig)
 static void
 install_handler (int sig_num, RETSIGTYPE (*sig_handler) (int sig))
 {
-#ifdef _POSIX_VERSION
+#ifdef SA_NOCLDSTOP
   struct sigaction sigact;
   sigaction (sig_num, NULL, &sigact);
   if (sigact.sa_handler != SIG_IGN)
@@ -430,15 +423,37 @@ install_handler (int sig_num, RETSIGTYPE (*sig_handler) (int sig))
 #endif
 }
 
+/* Open a file to a particular file descriptor.  This is like standard
+   `open', except it always returns DESIRED_FD if successful.  */
+static int
+open_fd (int desired_fd, char const *filename, int options, mode_t mode)
+{
+  int fd;
+  close (desired_fd);
+  fd = open (filename, options, mode);
+  if (fd < 0)
+    return -1;
+
+  if (fd != desired_fd)
+    {
+      if (dup2 (fd, desired_fd) != desired_fd)
+	desired_fd = -1;
+      if (close (fd) != 0)
+	return -1;
+    }
+
+  return desired_fd;
+}
+
 /* Write, then empty, the output buffer `obuf'. */
 
 static void
 write_output (void)
 {
-  int nwritten = full_write (output_fd, obuf, output_blocksize);
+  int nwritten = full_write (STDOUT_FILENO, obuf, output_blocksize);
   if (nwritten != output_blocksize)
     {
-      error (0, errno, "%s", output_file);
+      error (0, errno, _("writing to %s"), quote (output_file));
       if (nwritten > 0)
 	w_partial++;
       quit (1);
@@ -448,13 +463,14 @@ write_output (void)
   oc = 0;
 }
 
-/* Interpret one "conv=..." option. */
+/* Interpret one "conv=..." option.
+   As a by product, this function replaces each `,' in STR with a NUL byte.  */
 
 static void
 parse_conversion (char *str)
 {
   char *new;
-  int i;
+  unsigned int i;
 
   do
     {
@@ -469,7 +485,7 @@ parse_conversion (char *str)
 	  }
       if (conversions[i].convname == NULL)
 	{
-	  error (0, 0, _("%s: invalid conversion"), str);
+	  error (0, 0, _("invalid conversion: %s"), quote (str));
 	  usage (1);
 	}
       str = new;
@@ -481,101 +497,16 @@ parse_conversion (char *str)
    Assign nonzero to *INVALID if STR does not represent a number in
    this format. */
 
-/* FIXME: use xstrtou?[lq] */
-
 static uintmax_t
-parse_integer (char *str, int *invalid)
+parse_integer (const char *str, int *invalid)
 {
-  register uintmax_t n = 0;
-  register char *p = str;
+  uintmax_t n;
+  char *suffix;
+  enum strtol_error e = xstrtoumax (str, &suffix, 10, &n, "bcEGkMPTwYZ0");
 
-  while (ISDIGIT (*p))
+  if (e == LONGINT_INVALID_SUFFIX_CHAR && *suffix == 'x')
     {
-      uintmax_t n10 = n * 10;
-      int digit = *p - '0';
-      if (! (n10 / 10 == n && n10 <= n10 + digit))
-	{
-	  *invalid = 1;
-	  return 0;
-	}
-      n = n10 + digit;
-      p++;
-    }
-
-  for (;;)
-    {
-      uintmax_t multiplier;
-      int power = 0;
-
-#ifdef lint
-      /* Suppress `used before initialized' warning.  */
-      multiplier = 0;
-#endif
-
-      switch (*p++)
-	{
-	case '\0':
-	  return n;
-	case 'b':
-	  multiplier = 512;
-	  break;
-	case 'c':
-	  continue;
-	case 'E': /* Exa */
-	  power = 6;
-	  break;
-	case 'G': /* Giga */
-	  power = 3;
-	  break;
-	case 'k': /* kilo */
-	  power = 1;
-	  break;
-	case 'M': /* Mega */
-	  power = 2;
-	  break;
-	case 'P': /* Peta */
-	  power = 5;
-	  break;
-	case 'T': /* Tera */
-	  power = 4;
-	  break;
-	case 'w':
-	  multiplier = 2;
-	  break;
-	case 'x':
-	  multiplier = parse_integer (p, invalid);
-	  p = "";
-	  break;
-	case 'Y': /* Yotta */
-	  power = 8;
-	  break;
-	case 'Z': /* Zetta */
-	  power = 7;
-	  break;
-	default:
-	  {
-	    *invalid = 1;
-	    return 0;
-	  }
-	}
-
-      if (power)
-	{
-	  int base = 1024;
-
-	  switch (*p)
-	    {
-	    case 'B': p++; break;
-	    case 'D': p++; base = 1000; break;
-	    }
-
-	  for (multiplier = base;  --power;  multiplier *= base)
-	    if (multiplier * base / base != multiplier)
-	      {
-		*invalid = 1;
-		return 0;
-	      }
-	}
+      uintmax_t multiplier = parse_integer (suffix + 1, invalid);
 
       if (multiplier != 0 && n * multiplier / multiplier != n)
 	{
@@ -585,25 +516,22 @@ parse_integer (char *str, int *invalid)
 
       n *= multiplier;
     }
+  else if (e != LONGINT_OK)
+    {
+      *invalid = 1;
+      return 0;
+    }
+
+  return n;
 }
 
 static void
 scanargs (int argc, char **argv)
 {
   int i;
-  int c;
 
-  while ((c = getopt_long (argc, argv, "", long_options, NULL)) != -1)
-    {
-      switch (c)
-	{
-	case 0:
-	  break;
-
-	default:
-	  usage (1);
-	}
-    }
+  --argc;
+  ++argv;
 
   for (i = optind; i < argc; i++)
     {
@@ -613,7 +541,7 @@ scanargs (int argc, char **argv)
       val = strchr (name, '=');
       if (val == NULL)
 	{
-	  error (0, 0, _("unrecognized option `%s'"), name);
+	  error (0, 0, _("unrecognized option %s"), quote (name));
 	  usage (1);
 	}
       *val++ = '\0';
@@ -660,12 +588,13 @@ scanargs (int argc, char **argv)
 	    max_records = n;
 	  else
 	    {
-	      error (0, 0, _("unrecognized option `%s=%s'"), name, val);
+	      error (0, 0, _("unrecognized option %s=%s"),
+		     quote_n (0, name), quote_n (1, val));
 	      usage (1);
 	    }
 
 	  if (invalid)
-	    error (1, 0, _("invalid number `%s'"), val);
+	    error (1, 0, _("invalid number %s"), quote (val));
 	}
     }
 
@@ -687,7 +616,7 @@ scanargs (int argc, char **argv)
 static void
 apply_translations (void)
 {
-  int i;
+  unsigned int i;
 
 #define MX(a) (bit_count (conversions_mask & (a)))
   if ((MX (C_ASCII | C_EBCDIC | C_IBM) > 1)
@@ -707,14 +636,14 @@ only one conv in {ascii,ebcdic,ibm}, {lcase,ucase}, {block,unblock}, {unblock,sy
     {
       for (i = 0; i < 256; i++)
 	if (ISLOWER (trans_table[i]))
-	  trans_table[i] = toupper (trans_table[i]);
+	  trans_table[i] = TOUPPER (trans_table[i]);
       translation_needed = 1;
     }
   else if (conversions_mask & C_LCASE)
     {
       for (i = 0; i < 256; i++)
 	if (ISUPPER (trans_table[i]))
-	  trans_table[i] = tolower (trans_table[i]);
+	  trans_table[i] = TOLOWER (trans_table[i]);
       translation_needed = 1;
     }
 
@@ -736,10 +665,10 @@ only one conv in {ascii,ebcdic,ibm}, {lcase,ucase}, {block,unblock}, {unblock,sy
    to the NREAD bytes in BUF.  */
 
 static void
-translate_buffer (unsigned char *buf, int nread)
+translate_buffer (unsigned char *buf, size_t nread)
 {
-  register unsigned char *cp;
-  register int i;
+  unsigned char *cp;
+  size_t i;
 
   for (i = nread, cp = buf; i; i--, cp++)
     *cp = trans_table[*cp];
@@ -757,7 +686,7 @@ static unsigned char saved_char;
    next call.   Return the new start of the BUF buffer.  */
 
 static unsigned char *
-swab_buffer (unsigned char *buf, int *nread)
+swab_buffer (unsigned char *buf, size_t *nread)
 {
   unsigned char *bufstart = buf;
   register unsigned char *cp;
@@ -789,32 +718,56 @@ swab_buffer (unsigned char *buf, int *nread)
   return ++bufstart;
 }
 
+/* Return nonzero iff the file referenced by FDESC is of a type for
+   which lseek's return value is known to be invalid on some systems.
+   Otherwise, return zero.
+   For example, return nonzero if FDESC references a character device
+   (on any system) because the lseek on many Linux systems incorrectly
+   returns an offset implying it succeeds for tape devices, even though
+   the function fails to perform the requested operation.  In that case,
+   lseek should return nonzero and set errno.  */
+
+static int
+buggy_lseek_support (int fdesc)
+{
+  /* We have to resort to this because on some systems, lseek doesn't work
+     on some special files but doesn't return an error, either.
+     In particular, the Linux tape drivers are a problem.
+     For example, when I did the following using dd-4.0y or earlier on a
+     Linux-2.2.17 system with a Exabyte SCSI tape drive:
+
+       dev=/dev/nst0
+       reset='mt -f $dev rewind; mt -f $dev fsf 1'
+       eval $reset; dd if=$dev bs=32k of=out1
+       eval $reset; dd if=$dev bs=32k of=out2 skip=1
+
+     the resulting files, out1 and out2, would compare equal.  */
+
+  struct stat stats;
+
+  return (fstat (fdesc, &stats) == 0
+	  && (S_ISCHR (stats.st_mode)));
+}
+
 /* Throw away RECORDS blocks of BLOCKSIZE bytes on file descriptor FDESC,
    which is open with read permission for FILE.  Store up to BLOCKSIZE
-   bytes of the data at a time in BUF, if necessary. */
+   bytes of the data at a time in BUF, if necessary.  RECORDS must be
+   nonzero.  */
 
 static void
 skip (int fdesc, char *file, uintmax_t records, size_t blocksize,
       unsigned char *buf)
 {
-  struct stat stats;
   off_t o;
 
-  /* Use fstat instead of checking for errno == ESPIPE because
-     lseek doesn't work on some special files but doesn't return an
-     error, either. */
-  /* FIXME: can this really happen?  What system?  */
-  if (fstat (fdesc, &stats))
-    {
-      error (0, errno, "%s", file);
-      quit (1);
-    }
-
   /* Try lseek and if an error indicates it was an inappropriate
-     operation, fall back on using read.  */
+     operation, fall back on using read.  Some broken versions of
+     lseek may return zero, so count that as an error too as a valid
+     zero return is not possible here.  */
   o = records * blocksize;
   if (o / blocksize != records
-      || lseek (fdesc, o, SEEK_SET) == -1)
+      || buggy_lseek_support (fdesc)
+      || lseek (fdesc, o, SEEK_CUR) <= 0)
     {
       while (records-- > 0)
 	{
@@ -823,7 +776,7 @@ skip (int fdesc, char *file, uintmax_t records, size_t blocksize,
 	  nread = safe_read (fdesc, buf, blocksize);
 	  if (nread < 0)
 	    {
-	      error (0, errno, "%s", file);
+	      error (0, errno, _("reading %s"), quote (file));
 	      quit (1);
 	    }
 	  /* POSIX doesn't say what to do when dd detects it has been
@@ -865,9 +818,9 @@ copy_simple (unsigned char *buf, int nread)
    replacing the newline with trailing spaces).  */
 
 static void
-copy_with_block (unsigned char *buf, int nread)
+copy_with_block (unsigned char *buf, size_t nread)
 {
-  register int i;
+  size_t i;
 
   for (i = nread; i; i--, buf++)
     {
@@ -897,10 +850,10 @@ copy_with_block (unsigned char *buf, int nread)
    with a newline).  */
 
 static void
-copy_with_unblock (unsigned char *buf, int nread)
+copy_with_unblock (unsigned char *buf, size_t nread)
 {
-  register int i;
-  register unsigned char c;
+  size_t i;
+  unsigned char c;
   static int pending_spaces = 0;
 
   for (i = 0; i < nread; i++)
@@ -932,38 +885,67 @@ copy_with_unblock (unsigned char *buf, int nread)
 /* The main loop.  */
 
 static int
-copy (void)
+dd_copy (void)
 {
   unsigned char *ibuf, *bufstart; /* Input buffer. */
+  unsigned char *real_buf;	  /* real buffer address before alignment */
+  unsigned char *real_obuf;
   int nread;			/* Bytes read in the current block. */
   int exit_status = 0;
+  size_t page_size = getpagesize ();
+  size_t n_bytes_read;
 
   /* Leave at least one extra byte at the beginning and end of `ibuf'
      for conv=swab, but keep the buffer address even.  But some peculiar
      device drivers work only with word-aligned buffers, so leave an
      extra two bytes.  */
 
-  ibuf = (unsigned char *) xmalloc (input_blocksize + 2 * SWAB_ALIGN_OFFSET);
-  ibuf += SWAB_ALIGN_OFFSET;
+  /* Some devices require alignment on a sector or page boundary
+     (e.g. character disk devices).  Align the input buffer to a
+     page boundary to cover all bases.  Note that due to the swab
+     algorithm, we must have at least one byte in the page before
+     the input buffer;  thus we allocate 2 pages of slop in the
+     real buffer.  8k above the blocksize shouldn't bother anyone.
+
+     The page alignment is necessary on any linux system that supports
+     either the SGI raw I/O patch or Steven Tweedies raw I/O patch.
+     It is necessary when accessing raw (i.e. character special) disk
+     devices on Unixware or other SVR4-derived system.  */
+
+  real_buf = (unsigned char *) xmalloc (input_blocksize
+					+ 2 * SWAB_ALIGN_OFFSET
+					+ 2 * page_size - 1);
+  ibuf = real_buf;
+  ibuf += SWAB_ALIGN_OFFSET;	/* allow space for swab */
+
+  ibuf = PTR_ALIGN (ibuf, page_size);
 
   if (conversions_mask & C_TWOBUFS)
-    obuf = (unsigned char *) xmalloc (output_blocksize);
+    {
+      /* Page-align the output buffer, too.  */
+      real_obuf = (unsigned char *) xmalloc (output_blocksize + page_size - 1);
+      obuf = PTR_ALIGN (real_obuf, page_size);
+    }
   else
-    obuf = ibuf;
+    {
+      real_obuf = NULL;
+      obuf = ibuf;
+    }
 
   if (skip_records != 0)
-    skip (input_fd, input_file, skip_records, input_blocksize, ibuf);
+    skip (STDIN_FILENO, input_file, skip_records, input_blocksize, ibuf);
 
   if (seek_record != 0)
     {
       /* FIXME: this loses for
 	 % ./dd if=dd seek=1 |:
-	 ./dd: a1 standard output: Bad file number
+	 ./dd: standard output: Bad file number
 	 0+0 records in
 	 0+0 records out
 	 */
 
-	 skip (output_fd, output_file, seek_record, output_blocksize, obuf);
+	 skip (STDOUT_FILENO, output_file, seek_record, output_blocksize,
+	       obuf);
     }
 
   if (max_records == 0)
@@ -980,19 +962,19 @@ copy (void)
       if ((conversions_mask & C_SYNC) && (conversions_mask & C_NOERROR))
 	memset ((char *) ibuf, 0, input_blocksize);
 
-      nread = safe_read (input_fd, ibuf, input_blocksize);
+      nread = safe_read (STDIN_FILENO, ibuf, input_blocksize);
 
       if (nread == 0)
 	break;			/* EOF.  */
 
       if (nread < 0)
 	{
-	  error (0, errno, "%s", input_file);
+	  error (0, errno, _("reading %s"), quote (input_file));
 	  if (conversions_mask & C_NOERROR)
 	    {
 	      print_stats ();
 	      /* Seek past the bad block if possible. */
-	      lseek (input_fd, (off_t) input_blocksize, SEEK_CUR);
+	      lseek (STDIN_FILENO, (off_t) input_blocksize, SEEK_CUR);
 	      if (conversions_mask & C_SYNC)
 		/* Replace the missing input with null bytes and
 		   proceed normally.  */
@@ -1008,15 +990,18 @@ copy (void)
 	    }
 	}
 
-      if (nread < input_blocksize)
+      n_bytes_read = (size_t) nread;
+
+      if (n_bytes_read < input_blocksize)
 	{
 	  r_partial++;
 	  if (conversions_mask & C_SYNC)
 	    {
 	      if (!(conversions_mask & C_NOERROR))
 		/* If C_NOERROR, we zeroed the block before reading. */
-		memset ((char *) (ibuf + nread), 0, input_blocksize - nread);
-	      nread = input_blocksize;
+		memset ((char *) (ibuf + n_bytes_read), 0,
+			input_blocksize - n_bytes_read);
+	      n_bytes_read = input_blocksize;
 	    }
 	}
       else
@@ -1024,13 +1009,13 @@ copy (void)
 
       if (ibuf == obuf)		/* If not C_TWOBUFS. */
 	{
-	  int nwritten = full_write (output_fd, obuf, nread);
+	  int nwritten = full_write (STDOUT_FILENO, obuf, n_bytes_read);
 	  if (nwritten < 0)
 	    {
-	      error (0, errno, "%s", output_file);
+	      error (0, errno, _("writing %s"), quote (output_file));
 	      quit (1);
 	    }
-	  else if (nread == input_blocksize)
+	  else if (n_bytes_read == input_blocksize)
 	    w_full++;
 	  else
 	    w_partial++;
@@ -1040,19 +1025,19 @@ copy (void)
       /* Do any translations on the whole buffer at once.  */
 
       if (translation_needed)
-	translate_buffer (ibuf, nread);
+	translate_buffer (ibuf, n_bytes_read);
 
       if (conversions_mask & C_SWAB)
-	bufstart = swab_buffer (ibuf, &nread);
+	bufstart = swab_buffer (ibuf, &n_bytes_read);
       else
 	bufstart = ibuf;
 
       if (conversions_mask & C_BLOCK)
-        copy_with_block (bufstart, nread);
+        copy_with_block (bufstart, n_bytes_read);
       else if (conversions_mask & C_UNBLOCK)
-	copy_with_unblock (bufstart, nread);
+	copy_with_unblock (bufstart, n_bytes_read);
       else
-	copy_simple (bufstart, nread);
+	copy_simple (bufstart, n_bytes_read);
     }
 
   /* If we have a char left as a result of conv=swab, output it.  */
@@ -1070,7 +1055,7 @@ copy (void)
     {
       /* If the final input line didn't end with a '\n', pad
 	 the output block to `conversion_blocksize' chars.  */
-      size_t i;
+      unsigned int i;
       for (i = col; i < conversion_blocksize; i++)
 	output_char (space_character);
     }
@@ -1083,23 +1068,34 @@ copy (void)
   /* Write out the last block. */
   if (oc != 0)
     {
-      int nwritten = full_write (output_fd, obuf, oc);
+      int nwritten = full_write (STDOUT_FILENO, obuf, oc);
       if (nwritten > 0)
 	w_partial++;
       if (nwritten < 0)
 	{
-	  error (0, errno, "%s", output_file);
+	  error (0, errno, _("writing %s"), quote (output_file));
 	  quit (1);
 	}
     }
 
-  free (ibuf - SWAB_ALIGN_OFFSET);
-  if (obuf != ibuf)
-    free (obuf);
+  free (real_buf);
+  if (real_obuf)
+    free (real_obuf);
 
   return exit_status;
 }
 
+/* This is gross, but necessary, because of the way close_stdout
+   works and because this program closes STDOUT_FILENO directly.  */
+static void (*closeout_func) (void) = close_stdout;
+
+static void
+close_stdout_wrapper (void)
+{
+  if (closeout_func)
+    (*closeout_func) ();
+}
+
 int
 main (int argc, char **argv)
 {
@@ -1111,6 +1107,15 @@ main (int argc, char **argv)
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
+  /* Arrange to close stdout if parse_long_options exits.  */
+  atexit (close_stdout_wrapper);
+
+  parse_long_options (argc, argv, PROGRAM_NAME, GNU_PACKAGE, VERSION,
+		      AUTHORS, usage);
+
+  /* Don't close stdout on exit from here on.  */
+  closeout_func = NULL;
+
   /* Initialize translation table to identity translation. */
   for (i = 0; i < 256; i++)
     trans_table[i] = i;
@@ -1118,49 +1123,43 @@ main (int argc, char **argv)
   /* Decode arguments. */
   scanargs (argc, argv);
 
-  if (show_version)
-    {
-      printf ("dd (%s) %s\n", GNU_PACKAGE, VERSION);
-      close_stdout ();
-      exit (0);
-    }
-
-  if (show_help)
-    usage (0);
-
   apply_translations ();
 
   if (input_file != NULL)
     {
-      input_fd = open (input_file, O_RDONLY);
-      if (input_fd < 0)
-	error (1, errno, "%s", input_file);
+      if (open_fd (STDIN_FILENO, input_file, O_RDONLY, 0) < 0)
+	error (1, errno, _("opening %s"), quote (input_file));
     }
   else
     input_file = _("standard input");
 
-  if (input_fd == output_fd)
-    error (1, 0, _("%s is closed"), (input_fd == 0
-				     ? _("standard input")
-				     : _("standard output")));
-
   if (output_file != NULL)
     {
-      int omode = O_RDWR | O_CREAT;
+      mode_t perms = S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH;
+      int opts
+	= (O_CREAT
+	   | (seek_record || (conversions_mask & C_NOTRUNC) ? 0 : O_TRUNC));
 
-      if (seek_record == 0 && !(conversions_mask & C_NOTRUNC))
-	omode |= O_TRUNC;
-      output_fd = open (output_file, omode, 0666);
-      if (output_fd < 0)
-	error (1, errno, "%s", output_file);
+      /* Open the output file with *read* access only if we might
+	 need to read to satisfy a `seek=' request.  If we can't read
+	 the file, go ahead with write-only access; it might work.  */
+      if ((! seek_record
+	   || open_fd (STDOUT_FILENO, output_file, O_RDWR | opts, perms) < 0)
+	  && open_fd (STDOUT_FILENO, output_file, O_WRONLY | opts, perms) < 0)
+	error (1, errno, _("opening %s"), quote (output_file));
 #if HAVE_FTRUNCATE
       if (seek_record != 0 && !(conversions_mask & C_NOTRUNC))
 	{
 	  off_t o = seek_record * output_blocksize;
 	  if (o / output_blocksize != seek_record)
 	    error (1, 0, _("file offset out of range"));
-	  if (ftruncate (output_fd, o) < 0)
-	    error (0, errno, "%s", output_file);
+	  if (ftruncate (STDOUT_FILENO, o) < 0)
+	    {
+	      char buf[LONGEST_HUMAN_READABLE + 1];
+	      error (1, errno, _("advancing past %s blocks in output file %s"),
+				 human_readable (seek_record, buf, 1, 1),
+				 quote (output_file));
+	    }
 	}
 #endif
     }
@@ -1174,7 +1173,7 @@ main (int argc, char **argv)
   install_handler (SIGPIPE, interrupt_handler);
   install_handler (SIGINFO, siginfo_handler);
 
-  exit_status = copy ();
+  exit_status = dd_copy ();
 
   quit (exit_status);
 }
