@@ -1,5 +1,5 @@
 /* du -- summarize disk usage
-   Copyright (C) 88, 89, 90, 91, 95, 96, 97, 1998 Free Software Foundation, Inc.
+   Copyright (C) 88, 89, 90, 91, 1995-2000 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -51,14 +51,20 @@
 #include <sys/types.h>
 #include <assert.h>
 
-#include "exclude.h"
 #include "system.h"
-#include "save-cwd.h"
-#include "closeout.h"
 #include "error.h"
+#include "exclude.h"
 #include "human.h"
-#include "xstrtol.h"
+#include "quote.h"
+#include "save-cwd.h"
 #include "savedir.h"
+#include "xstrtol.h"
+
+/* The official name of this program (e.g., no `g' prefix).  */
+#define PROGRAM_NAME "du"
+
+#define AUTHORS \
+  "Torbjorn Granlund, David MacKenzie, Larry McVoy, and Paul Eggert"
 
 /* Initial number of entries in each hash table entry's table of inodes.  */
 #define INITIAL_HASH_MODULE 100
@@ -104,19 +110,6 @@ typedef struct String String;
 int stat ();
 int lstat ();
 
-static int hash_insert PARAMS ((ino_t ino, dev_t dev));
-static int hash_insert2 PARAMS ((struct htab *_htab, ino_t ino, dev_t dev));
-static uintmax_t count_entry PARAMS ((const char *ent, int top, dev_t last_dev,
-				      int depth));
-static void du_files PARAMS ((char **files));
-static void hash_init PARAMS ((unsigned int modulus,
-			       unsigned int entry_tab_size));
-static void hash_reset PARAMS ((void));
-static void str_concatc PARAMS ((String *s1, char *cstr));
-static void str_copyc PARAMS ((String *s1, char *cstr));
-static void str_init PARAMS ((String **s1, unsigned int size));
-static void str_trunc PARAMS ((String *s1, unsigned int length));
-
 /* Name under which this program was invoked.  */
 char *program_name;
 
@@ -130,7 +123,7 @@ static int opt_count_all = 0;
 static int opt_one_file_system = 0;
 
 /* If nonzero, print a grand total at the end. */
-static int opt_combined_arguments = 0;
+static int print_totals = 0;
 
 /* If nonzero, do not add sizes of subdirectories. */
 static int opt_separate_dirs = 0;
@@ -153,9 +146,6 @@ static String *path;
 /* Pointer to hash structure, used by the hash routines.  */
 static struct htab *htab;
 
-/* Globally used stat buffer.  */
-static struct stat stat_buf;
-
 /* A pointer to either lstat or stat, depending on whether
    dereferencing of all symbolic links is to be done. */
 static int (*xstat) ();
@@ -163,49 +153,48 @@ static int (*xstat) ();
 /* The exit status to use if we don't get any fatal errors. */
 static int exit_status;
 
-/* If nonzero, display usage information and exit.  */
-static int show_help;
-
-/* If nonzero, print the version on standard output and exit.  */
-static int show_version;
-
 /* File name patterns to exclude.  */
 static struct exclude *exclude;
 
 /* Grand total size of all args, in units of ST_NBLOCKSIZE-byte blocks. */
 static uintmax_t tot_size = 0;
 
+/* For long options that have no equivalent short option, use a
+   non-character as a pseudo short option, starting with CHAR_MAX + 1.  */
+enum
+{
+  EXCLUDE_OPTION = CHAR_MAX + 1,
+  BLOCK_SIZE_OPTION,
+  MAX_DEPTH_OPTION
+};
+
 static struct option const long_options[] =
 {
   {"all", no_argument, NULL, 'a'},
-  {"block-size", required_argument, 0, CHAR_MAX + 2},
+  {"block-size", required_argument, 0, BLOCK_SIZE_OPTION},
   {"bytes", no_argument, NULL, 'b'},
   {"count-links", no_argument, NULL, 'l'},
   {"dereference", no_argument, NULL, 'L'},
   {"dereference-args", no_argument, NULL, 'D'},
-  {"exclude", required_argument, 0, CHAR_MAX + 1},
+  {"exclude", required_argument, 0, EXCLUDE_OPTION},
   {"exclude-from", required_argument, 0, 'X'},
   {"human-readable", no_argument, NULL, 'h'},
   {"si", no_argument, 0, 'H'},
   {"kilobytes", no_argument, NULL, 'k'},
-  {"max-depth", required_argument, NULL, CHAR_MAX + 3},
+  {"max-depth", required_argument, NULL, MAX_DEPTH_OPTION},
   {"megabytes", no_argument, NULL, 'm'},
   {"one-file-system", no_argument, NULL, 'x'},
   {"separate-dirs", no_argument, NULL, 'S'},
   {"summarize", no_argument, NULL, 's'},
   {"total", no_argument, NULL, 'c'},
-  {"help", no_argument, &show_help, 1},
-  {"version", no_argument, &show_version, 1},
+  {GETOPT_HELP_OPTION_DECL},
+  {GETOPT_VERSION_OPTION_DECL},
   {NULL, 0, NULL, 0}
 };
 
-static void
-usage (int status, char *reason)
+void
+usage (int status)
 {
-  if (reason != NULL)
-    fprintf (status == 0 ? stdout : stderr, "%s: %s\n",
-	     program_name, reason);
-
   if (status != 0)
     fprintf (stderr, _("Try `%s --help' for more information.\n"),
 	     program_name);
@@ -239,161 +228,67 @@ Summarize disk usage of each FILE, recursively for directories.\n\
       --version         output version information and exit\n\
 "));
       puts (_("\nReport bugs to <bug-fileutils@gnu.org>."));
-      close_stdout ();
     }
   exit (status);
 }
 
-int
-main (int argc, char **argv)
+/* Initialize string S1 to hold SIZE characters.  */
+
+static void
+str_init (String **s1, unsigned int size)
 {
-  int c;
-  char *cwd_only[2];
-  int max_depth_specified = 0;
+  String *s;
 
-  /* If nonzero, display only a total for each argument. */
-  int opt_summarize_only = 0;
+  s = (String *) xmalloc (sizeof (struct String));
+  s->text = xmalloc (size + 1);
 
-  cwd_only[0] = ".";
-  cwd_only[1] = NULL;
+  s->alloc = size;
+  *s1 = s;
+}
 
-  program_name = argv[0];
-  setlocale (LC_ALL, "");
-  bindtextdomain (PACKAGE, LOCALEDIR);
-  textdomain (PACKAGE);
-
-  exclude = new_exclude ();
-  xstat = lstat;
-
-  human_block_size (getenv ("DU_BLOCK_SIZE"), 0, &output_block_size);
-
-  while ((c = getopt_long (argc, argv, "abchHklmsxDLSX:", long_options, NULL))
-	 != -1)
+static void
+ensure_space (String *s, unsigned int size)
+{
+  if (s->alloc < size)
     {
-      long int tmp_long;
-      switch (c)
-	{
-	case 0:			/* Long option. */
-	  break;
-
-	case 'a':
-	  opt_all = 1;
-	  break;
-
-	case 'b':
-	  output_block_size = 1;
-	  break;
-
-	case 'c':
-	  opt_combined_arguments = 1;
-	  break;
-
-	case 'h':
-	  output_block_size = -1024;
-	  break;
-
-	case 'H':
-	  output_block_size = -1000;
-	  break;
-
-	case 'k':
-	  output_block_size = 1024;
-	  break;
-
-	case CHAR_MAX + 3:		/* --max-depth=N */
-	  if (xstrtol (optarg, NULL, 0, &tmp_long, NULL) != LONGINT_OK
-	      || tmp_long < 0 || tmp_long > INT_MAX)
-	    error (1, 0, _("invalid maximum depth `%s'"), optarg);
-
-	  max_depth_specified = 1;
-	  max_depth = (int) tmp_long;
- 	  break;
-
-	case 'm':
-	  output_block_size = 1024 * 1024;
-	  break;
-
-	case 'l':
-	  opt_count_all = 1;
-	  break;
-
-	case 's':
-	  opt_summarize_only = 1;
-	  break;
-
-	case 'x':
-	  opt_one_file_system = 1;
-	  break;
-
-	case 'D':
-	  opt_dereference_arguments = 1;
-	  break;
-
-	case 'L':
-	  xstat = stat;
-	  break;
-
-	case 'S':
-	  opt_separate_dirs = 1;
-	  break;
-
-	case 'X':
-	  if (add_exclude_file (exclude, optarg, '\n') != 0)
-	    error (1, errno, "%s", optarg);
-	  break;
-
-	case CHAR_MAX + 1:
-	  add_exclude (exclude, optarg);
-	  break;
-
-	case CHAR_MAX + 2:
-	  human_block_size (optarg, 1, &output_block_size);
-	  break;
-
-	default:
-	  usage (1, (char *) 0);
-	}
+      s->text = xrealloc (s->text, size + 1);
+      s->alloc = size;
     }
+}
 
-  if (show_version)
+/* Assign the null-terminated C-string CSTR to S1.  */
+
+static void
+str_copyc (String *s1, const char *cstr)
+{
+  unsigned l = strlen (cstr);
+  ensure_space (s1, l);
+  strcpy (s1->text, cstr);
+  s1->length = l;
+}
+
+static void
+str_concatc (String *s1, const char *cstr)
+{
+  unsigned l1 = s1->length;
+  unsigned l2 = strlen (cstr);
+  unsigned l = l1 + l2;
+
+  ensure_space (s1, l);
+  strcpy (s1->text + l1, cstr);
+  s1->length = l;
+}
+
+/* Truncate the string S1 to have length LENGTH.  */
+
+static void
+str_trunc (String *s1, unsigned int length)
+{
+  if (s1->length > length)
     {
-      printf ("du (%s) %s\n", GNU_PACKAGE, VERSION);
-      close_stdout ();
-      exit (0);
+      s1->text[length] = 0;
+      s1->length = length;
     }
-
-  if (show_help)
-    usage (0, NULL);
-
-  if (opt_all && opt_summarize_only)
-    usage (1, _("cannot both summarize and show all entries"));
-
-  if (opt_summarize_only && max_depth_specified && max_depth == 0)
-    {
-      error (0, 0,
-	     _("warning: summarizing is the same as using --max-depth=0"));
-    }
-
-  if (opt_summarize_only && max_depth_specified && max_depth != 0)
-    {
-      error (0, 0,
-	     _("warning: summarizing conflicts with --max-depth=%d"),
-	       max_depth);
-      usage (1, NULL);
-    }
-
-  if (opt_summarize_only)
-    max_depth = 0;
-
-  /* Initialize the hash structure for inode numbers.  */
-  hash_init (INITIAL_HASH_MODULE, INITIAL_ENTRY_TAB_SIZE);
-
-  str_init (&path, INITIAL_PATH_SIZE);
-
-  du_files (optind == argc ? cwd_only : argv + optind);
-
-  close_stdout ();
-  exit (exit_status);
 }
 
 /* Print N_BLOCKS followed by STRING on a line.  NBLOCKS is the number of
@@ -406,200 +301,26 @@ print_size (uintmax_t n_blocks, const char *string)
 {
   char buf[LONGEST_HUMAN_READABLE + 1];
   printf ("%s\t%s\n",
-	  human_readable (n_blocks, buf, ST_NBLOCKSIZE, output_block_size),
+	  human_readable_inexact (n_blocks, buf, ST_NBLOCKSIZE,
+				  output_block_size, human_ceiling),
 	  string);
   fflush (stdout);
 }
 
-/* Recursively print the sizes of the directories (and, if selected, files)
-   named in FILES, the last entry of which is NULL.  */
+/* Reset the hash structure in the global variable `htab' to
+   contain no entries.  */
 
 static void
-du_files (char **files)
+hash_reset (void)
 {
-  struct saved_cwd cwd;
-  ino_t initial_ino;		/* Initial directory's inode. */
-  dev_t initial_dev;		/* Initial directory's device. */
-  int i;			/* Index in FILES. */
+  int i;
+  struct entry **p;
 
-  if (save_cwd (&cwd))
-    exit (1);
+  htab->first_free_entry = 0;
 
-  /* Remember the inode and device number of the current directory.  */
-  if (stat (".", &stat_buf))
-    error (1, errno, _("current directory"));
-  initial_ino = stat_buf.st_ino;
-  initial_dev = stat_buf.st_dev;
-
-  for (i = 0; files[i]; i++)
-    {
-      char *arg;
-      int s;
-
-      arg = files[i];
-
-      /* Delete final slash in the argument, unless the slash is alone.  */
-      s = strlen (arg) - 1;
-      if (s != 0)
-	{
-	  if (arg[s] == '/')
-	    arg[s] = 0;
-
-	  str_copyc (path, arg);
-	}
-      else if (arg[0] == '/')
-	str_trunc (path, 0);	/* Null path for root directory.  */
-      else
-	str_copyc (path, arg);
-
-      if (!opt_combined_arguments)
-	hash_reset ();
-
-      count_entry (arg, 1, 0, 0);
-
-      /* chdir if `count_entry' has changed the working directory.  */
-      if (stat (".", &stat_buf))
-	error (1, errno, ".");
-      if (stat_buf.st_ino != initial_ino || stat_buf.st_dev != initial_dev)
-	{
-	  if (restore_cwd (&cwd, _("starting directory"), NULL))
-	    exit (1);
-	}
-    }
-
-  if (opt_combined_arguments)
-    print_size (tot_size, _("total"));
-
-  free_cwd (&cwd);
-}
-
-/* Print (if appropriate) the size (in units determined by `output_block_size')
-   of file or directory ENT. Return the size of ENT in units of 512-byte
-   blocks.  TOP is one for external calls, zero for recursive calls.
-   LAST_DEV is the device that the parent directory of ENT is on.
-   DEPTH is the number of levels (in hierarchy) down from a command
-   line argument.  Don't print if DEPTH > max_depth.  */
-
-static uintmax_t
-count_entry (const char *ent, int top, dev_t last_dev, int depth)
-{
-  uintmax_t size;
-
-  if (((top && opt_dereference_arguments)
-       ? stat (ent, &stat_buf)
-       : (*xstat) (ent, &stat_buf)) < 0)
-    {
-      error (0, errno, "%s", path->text);
-      exit_status = 1;
-      return 0;
-    }
-
-  if (!opt_count_all
-      && stat_buf.st_nlink > 1
-      && hash_insert (stat_buf.st_ino, stat_buf.st_dev))
-    return 0;			/* Have counted this already.  */
-
-  size = ST_NBLOCKS (stat_buf);
-  tot_size += size;
-
-  if (S_ISDIR (stat_buf.st_mode))
-    {
-      unsigned pathlen;
-      dev_t dir_dev;
-      char *name_space;
-      char *namep;
-      struct saved_cwd cwd;
-      int through_symlink;
-      struct stat e_buf;
-
-      dir_dev = stat_buf.st_dev;
-
-      if (opt_one_file_system && !top && last_dev != dir_dev)
-	return 0;		/* Don't enter a new file system.  */
-
-#ifndef S_ISLNK
-# define S_ISLNK(s) 0
-#endif
-      /* If we're dereferencing symlinks and we're about to chdir through
-	 a symlink, remember the current directory so we can return to it
-	 later.  In other cases, chdir ("..") works fine.  */
-      through_symlink = (xstat == stat
-			 && lstat (ent, &e_buf) == 0
-			 && S_ISLNK (e_buf.st_mode));
-      if (through_symlink && save_cwd (&cwd))
-	exit (1);
-
-      if (chdir (ent) < 0)
-	{
-	  error (0, errno, _("cannot change to directory %s"), path->text);
-	  exit_status = 1;
-	  return 0;
-	}
-
-      errno = 0;
-      name_space = savedir (".", (unsigned int) stat_buf.st_size);
-      if (name_space == NULL)
-	{
-	  if (errno)
-	    {
-	      error (0, errno, "%s", path->text);
-	      if (through_symlink)
-		{
-		  if (restore_cwd (&cwd, "..", path->text))
-		    exit (1);
-		  free_cwd (&cwd);
-		}
-	      else if (chdir ("..") < 0)
-		  error (1, errno, _("cannot change to `..' from directory %s"),
-			 path->text);
-	      exit_status = 1;
-	      return 0;
-	    }
-	  else
-	    error (1, 0, _("virtual memory exhausted"));
-	}
-
-      /* Remember the current path.  */
-
-      str_concatc (path, "/");
-      pathlen = path->length;
-
-      for (namep = name_space; *namep; namep += strlen (namep) + 1)
-	{
-	  if (!excluded_filename (exclude, namep))
-	    {
-	      str_concatc (path, namep);
-	      size += count_entry (namep, 0, dir_dev, depth + 1);
-	      str_trunc (path, pathlen);
-	    }
-	}
-
-      free (name_space);
-      if (through_symlink)
-	{
-	  restore_cwd (&cwd, "..", path->text);
-	  free_cwd (&cwd);
-	}
-      else if (chdir ("..") < 0)
-	{
-	  error (1, errno,
-		 _("cannot change to `..' from directory %s"), path->text);
-	}
-
-      str_trunc (path, pathlen - 1); /* Remove the "/" we added.  */
-      if (depth <= max_depth || top)
-	print_size (size, path->length > 0 ? path->text : "/");
-      return opt_separate_dirs ? 0 : size;
-    }
-  else if ((opt_all && depth <= max_depth) || top)
-    {
-      /* FIXME: make this an option.  */
-      int print_only_dir_size = 0;
-      if (!print_only_dir_size)
-	print_size (size, path->length > 0 ? path->text : "/");
-    }
-
-  return size;
+  p = htab->hash;
+  for (i = htab->modulus; i > 0; i--)
+    *p++ = NULL;
 }
 
 /* Allocate space for the hash structures, and set the global
@@ -627,20 +348,43 @@ hash_init (unsigned int modulus, unsigned int entry_tab_size)
   hash_reset ();
 }
 
-/* Reset the hash structure in the global variable `htab' to
-   contain no entries.  */
+/* Insert INO and DEV in the hash structure HTAB, if not
+   already present.  Return zero if inserted and nonzero if it
+   already existed.  */
 
-static void
-hash_reset (void)
+static int
+hash_insert2 (struct htab *ht, ino_t ino, dev_t dev)
 {
-  int i;
-  struct entry **p;
+  struct entry **hp, *ep2, *ep;
+  hp = &ht->hash[ino % ht->modulus];
+  ep2 = *hp;
 
-  htab->first_free_entry = 0;
+  /* Collision?  */
 
-  p = htab->hash;
-  for (i = htab->modulus; i > 0; i--)
-    *p++ = NULL;
+  if (ep2 != NULL)
+    {
+      ep = ep2;
+
+      /* Search for an entry with the same data.  */
+
+      do
+	{
+	  if (ep->ino == ino && ep->dev == dev)
+	    return 1;		/* Found an entry with the same data.  */
+	  ep = ep->coll_link;
+	}
+      while (ep != NULL);
+
+      /* Did not find it.  */
+
+    }
+
+  ep = *hp = &ht->entry_tab[ht->first_free_entry++];
+  ep->ino = ino;
+  ep->dev = dev;
+  ep->coll_link = ep2;		/* `ep2' is NULL if no collision.  */
+
+  return 0;
 }
 
 /* Insert an item (inode INO and device DEV) in the hash
@@ -711,100 +455,335 @@ hash_insert (ino_t ino, dev_t dev)
   return hash_insert2 (htab_r, ino, dev);
 }
 
-/* Insert INO and DEV in the hash structure HTAB, if not
-   already present.  Return zero if inserted and nonzero if it
-   already existed.  */
+/* Restore the previous working directory or exit.
+   If CWD is null, simply call `chdir ("..")'.  Otherwise,
+   use CWD and free it.  CURR_DIR_NAME is the name of the current directory
+   and is used solely in failure diagnostics.  */
 
-static int
-hash_insert2 (struct htab *ht, ino_t ino, dev_t dev)
+static void
+pop_dir (struct saved_cwd *cwd, const char *curr_dir_name)
 {
-  struct entry **hp, *ep2, *ep;
-  hp = &ht->hash[ino % ht->modulus];
-  ep2 = *hp;
-
-  /* Collision?  */
-
-  if (ep2 != NULL)
+  if (cwd)
     {
-      ep = ep2;
+      if (restore_cwd (cwd, "..", curr_dir_name))
+	exit (1);
+      free_cwd (cwd);
+    }
+  else if (chdir ("..") < 0)
+    {
+      error (1, errno, _("cannot change to `..' from directory %s"),
+	     quote (curr_dir_name));
+    }
+}
 
-      /* Search for an entry with the same data.  */
+/* Print (if appropriate) the size (in units determined by `output_block_size')
+   of file or directory ENT. Return the size of ENT in units of 512-byte
+   blocks.  TOP is one for external calls, zero for recursive calls.
+   LAST_DEV is the device that the parent directory of ENT is on.
+   DEPTH is the number of levels (in hierarchy) down from a command
+   line argument.  Don't print if DEPTH > max_depth.
+   An important invariant is that when this function returns, the current
+   working directory is the same as when it was called.  */
 
-      do
+static uintmax_t
+count_entry (const char *ent, int top, dev_t last_dev, int depth)
+{
+  uintmax_t size;
+  struct stat stat_buf;
+
+  if (((top && opt_dereference_arguments)
+       ? stat (ent, &stat_buf)
+       : (*xstat) (ent, &stat_buf)) < 0)
+    {
+      error (0, errno, "%s", quote (path->text));
+      exit_status = 1;
+      return 0;
+    }
+
+  if (!opt_count_all
+      && stat_buf.st_nlink > 1
+      && hash_insert (stat_buf.st_ino, stat_buf.st_dev))
+    return 0;			/* Have counted this already.  */
+
+  size = ST_NBLOCKS (stat_buf);
+  tot_size += size;
+
+  if (S_ISDIR (stat_buf.st_mode))
+    {
+      unsigned pathlen;
+      dev_t dir_dev;
+      char *name_space;
+      char *namep;
+      struct saved_cwd *cwd;
+      struct saved_cwd cwd_buf;
+      struct stat e_buf;
+
+      dir_dev = stat_buf.st_dev;
+
+      if (opt_one_file_system && !top && last_dev != dir_dev)
+	return 0;		/* Don't enter a new file system.  */
+
+#ifndef S_ISLNK
+# define S_ISLNK(s) 0
+#endif
+      /* If we're traversing more than one level, or if we're
+	 dereferencing symlinks and we're about to chdir through a
+	 symlink, remember the current directory so we can return to
+	 it later.  In other cases, chdir ("..") works fine.
+	 Treat `.' and `..' like multi-level paths, since `chdir ("..")'
+	 wont't restore the current working directory after a `chdir'
+	 to one of those.  */
+      if (strchr (ent, '/')
+	  || DOT_OR_DOTDOT (ent)
+	  || (xstat == stat
+	      && lstat (ent, &e_buf) == 0
+	      && S_ISLNK (e_buf.st_mode)))
 	{
-	  if (ep->ino == ino && ep->dev == dev)
-	    return 1;		/* Found an entry with the same data.  */
-	  ep = ep->coll_link;
+	  if (save_cwd (&cwd_buf))
+	    exit (1);
+	  cwd = &cwd_buf;
 	}
-      while (ep != NULL);
+      else
+	cwd = NULL;
 
-      /* Did not find it.  */
+      if (chdir (ent) < 0)
+	{
+	  error (0, errno, _("cannot change to directory %s"),
+		 quote (path->text));
+	  if (cwd)
+	    free_cwd (cwd);
+	  exit_status = 1;
+	  return 0;
+	}
 
+      name_space = savedir (".", stat_buf.st_size);
+      if (name_space == NULL)
+	{
+	  error (0, errno, "%s", quote (path->text));
+	  pop_dir (cwd, path->text);
+	  exit_status = 1;
+	  return 0;
+	}
+
+      /* Remember the current path.  */
+
+      str_concatc (path, "/");
+      pathlen = path->length;
+
+      for (namep = name_space; *namep; namep += strlen (namep) + 1)
+	{
+	  if (!excluded_filename (exclude, namep, 0))
+	    {
+	      str_concatc (path, namep);
+	      size += count_entry (namep, 0, dir_dev, depth + 1);
+	      str_trunc (path, pathlen);
+	    }
+	}
+
+      free (name_space);
+      pop_dir (cwd, path->text);
+
+      str_trunc (path, pathlen - 1); /* Remove the "/" we added.  */
+      if (depth <= max_depth || top)
+	print_size (size, path->length > 0 ? path->text : "/");
+      return opt_separate_dirs ? 0 : size;
     }
-
-  ep = *hp = &ht->entry_tab[ht->first_free_entry++];
-  ep->ino = ino;
-  ep->dev = dev;
-  ep->coll_link = ep2;		/* `ep2' is NULL if no collision.  */
-
-  return 0;
-}
-
-/* Initialize string S1 to hold SIZE characters.  */
-
-static void
-str_init (String **s1, unsigned int size)
-{
-  String *s;
-
-  s = (String *) xmalloc (sizeof (struct String));
-  s->text = xmalloc (size + 1);
-
-  s->alloc = size;
-  *s1 = s;
-}
-
-static void
-ensure_space (String *s, unsigned int size)
-{
-  if (s->alloc < size)
+  else if ((opt_all && depth <= max_depth) || top)
     {
-      s->text = xrealloc (s->text, size + 1);
-      s->alloc = size;
+      /* FIXME: make this an option.  */
+      int print_only_dir_size = 0;
+      if (!print_only_dir_size)
+	print_size (size, path->length > 0 ? path->text : "/");
     }
+
+  return size;
 }
 
-/* Assign the null-terminated C-string CSTR to S1.  */
+/* Recursively print the sizes of the directories (and, if selected, files)
+   named in FILES, the last entry of which is NULL.  */
 
 static void
-str_copyc (String *s1, char *cstr)
+du_files (char **files)
 {
-  unsigned l = strlen (cstr);
-  ensure_space (s1, l);
-  strcpy (s1->text, cstr);
-  s1->length = l;
-}
+  int i;			/* Index in FILES. */
 
-static void
-str_concatc (String *s1, char *cstr)
-{
-  unsigned l1 = s1->length;
-  unsigned l2 = strlen (cstr);
-  unsigned l = l1 + l2;
-
-  ensure_space (s1, l);
-  strcpy (s1->text + l1, cstr);
-  s1->length = l;
-}
-
-/* Truncate the string S1 to have length LENGTH.  */
-
-static void
-str_trunc (String *s1, unsigned int length)
-{
-  if (s1->length > length)
+  for (i = 0; files[i]; i++)
     {
-      s1->text[length] = 0;
-      s1->length = length;
+      char *arg;
+      int s;
+
+      arg = files[i];
+
+      /* Delete final slash in the argument, unless the slash is alone.  */
+      s = strlen (arg) - 1;
+      if (s != 0)
+	{
+	  if (arg[s] == '/')
+	    arg[s] = 0;
+
+	  str_copyc (path, arg);
+	}
+      else if (arg[0] == '/')
+	str_trunc (path, 0);	/* Null path for root directory.  */
+      else
+	str_copyc (path, arg);
+
+      if (!print_totals)
+	hash_reset ();
+
+      count_entry (arg, 1, 0, 0);
     }
+
+  if (print_totals)
+    print_size (tot_size, _("total"));
+}
+
+int
+main (int argc, char **argv)
+{
+  int c;
+  char *cwd_only[2];
+  int max_depth_specified = 0;
+
+  /* If nonzero, display only a total for each argument. */
+  int opt_summarize_only = 0;
+
+  cwd_only[0] = ".";
+  cwd_only[1] = NULL;
+
+  program_name = argv[0];
+  setlocale (LC_ALL, "");
+  bindtextdomain (PACKAGE, LOCALEDIR);
+  textdomain (PACKAGE);
+
+  atexit (close_stdout);
+
+  exclude = new_exclude ();
+  xstat = lstat;
+
+  human_block_size (getenv ("DU_BLOCK_SIZE"), 0, &output_block_size);
+
+  while ((c = getopt_long (argc, argv, "abchHklmsxDLSX:", long_options, NULL))
+	 != -1)
+    {
+      long int tmp_long;
+      switch (c)
+	{
+	case 0:			/* Long option. */
+	  break;
+
+	case 'a':
+	  opt_all = 1;
+	  break;
+
+	case 'b':
+	  output_block_size = 1;
+	  break;
+
+	case 'c':
+	  print_totals = 1;
+	  break;
+
+	case 'h':
+	  output_block_size = -1024;
+	  break;
+
+	case 'H':
+	  output_block_size = -1000;
+	  break;
+
+	case 'k':
+	  output_block_size = 1024;
+	  break;
+
+	case MAX_DEPTH_OPTION:		/* --max-depth=N */
+	  if (xstrtol (optarg, NULL, 0, &tmp_long, NULL) != LONGINT_OK
+	      || tmp_long < 0 || tmp_long > INT_MAX)
+	    error (1, 0, _("invalid maximum depth %s"), quote (optarg));
+
+	  max_depth_specified = 1;
+	  max_depth = (int) tmp_long;
+ 	  break;
+
+	case 'm':
+	  output_block_size = 1024 * 1024;
+	  break;
+
+	case 'l':
+	  opt_count_all = 1;
+	  break;
+
+	case 's':
+	  opt_summarize_only = 1;
+	  break;
+
+	case 'x':
+	  opt_one_file_system = 1;
+	  break;
+
+	case 'D':
+	  opt_dereference_arguments = 1;
+	  break;
+
+	case 'L':
+	  xstat = stat;
+	  break;
+
+	case 'S':
+	  opt_separate_dirs = 1;
+	  break;
+
+	case 'X':
+	  if (add_exclude_file (add_exclude, exclude, optarg, '\n') != 0)
+	    error (1, errno, "%s", quote (optarg));
+	  break;
+
+	case EXCLUDE_OPTION:
+	  add_exclude (exclude, optarg);
+	  break;
+
+	case BLOCK_SIZE_OPTION:
+	  human_block_size (optarg, 1, &output_block_size);
+	  break;
+
+	case_GETOPT_HELP_CHAR;
+
+	case_GETOPT_VERSION_CHAR (PROGRAM_NAME, AUTHORS);
+
+	default:
+	  usage (1);
+	}
+    }
+
+  if (opt_all && opt_summarize_only)
+    {
+      error (0, 0, _("cannot both summarize and show all entries"));
+      usage (1);
+    }
+
+  if (opt_summarize_only && max_depth_specified && max_depth == 0)
+    {
+      error (0, 0,
+	     _("warning: summarizing is the same as using --max-depth=0"));
+    }
+
+  if (opt_summarize_only && max_depth_specified && max_depth != 0)
+    {
+      error (0, 0,
+	     _("warning: summarizing conflicts with --max-depth=%d"),
+	       max_depth);
+      usage (1);
+    }
+
+  if (opt_summarize_only)
+    max_depth = 0;
+
+  /* Initialize the hash structure for inode numbers.  */
+  hash_init (INITIAL_HASH_MODULE, INITIAL_ENTRY_TAB_SIZE);
+
+  str_init (&path, INITIAL_PATH_SIZE);
+
+  du_files (optind == argc ? cwd_only : argv + optind);
+
+  exit (exit_status);
 }
