@@ -1,5 +1,5 @@
 /* install - copy files and set attributes
-   Copyright (C) 89, 90, 91, 95, 96, 97, 1998 Free Software Foundation, Inc.
+   Copyright (C) 89, 90, 91, 1995-2000 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -30,7 +30,7 @@
 
    -m, --mode=MODE
 	Set the permission mode for the installed file or directory
-	to MODE, which is an octal number (default is 0755).
+	to MODE, which is an octal number (default is u=rwx,g=rx,o=rx).
 
    -o, --owner=OWNER
 	If run as root, set the ownership of the installed file to
@@ -73,14 +73,20 @@
 
 #include "system.h"
 #include "backupfile.h"
-#include "modechange.h"
-#include "makepath.h"
-#include "closeout.h"
 #include "error.h"
-#include "xstrtol.h"
-#include "path-concat.h"
 #include "cp-hash.h"
 #include "copy.h"
+#include "dirname.h"
+#include "makepath.h"
+#include "modechange.h"
+#include "path-concat.h"
+#include "quote.h"
+#include "xstrtol.h"
+
+/* The official name of this program (e.g., no `g' prefix).  */
+#define PROGRAM_NAME "install"
+
+#define AUTHORS "David MacKenzie"
 
 #if HAVE_SYS_WAIT_H
 # include <sys/wait.h>
@@ -112,25 +118,11 @@ gid_t getgid ();
 /* Initial number of entries in the inode hash table.  */
 #define INITIAL_ENTRY_TAB_SIZE 70
 
-/* True if C is an ASCII octal digit. */
-#define isodigit(c) ((c) >= '0' && c <= '7')
-
 /* Number of bytes of a file to copy at a time. */
 #define READ_SIZE (32 * 1024)
 
-#ifndef UID_T_MAX
-# define UID_T_MAX TYPE_MAXIMUM (uid_t)
-#endif
-
-#ifndef GID_T_MAX
-# define GID_T_MAX TYPE_MAXIMUM (gid_t)
-#endif
-
-char *base_name ();
-char *dirname ();
 int full_write ();
 int isdir ();
-enum backup_type get_version ();
 
 int stat ();
 
@@ -146,14 +138,11 @@ static int install_file_in_file PARAMS ((const char *from, const char *to,
 					 const struct cp_options *x));
 static void get_ids PARAMS ((void));
 static void strip PARAMS ((const char *path));
-static void usage PARAMS ((int status));
+void usage PARAMS ((int status));
 
 /* The name this program was run with, for error messages. */
 char *program_name;
 
-/* FIXME: document */
-enum backup_type backup_type;
-
 /* The user name that will own the files, or NULL to make the owner
    the current user ID. */
 static char *owner_name;
@@ -170,7 +159,7 @@ static gid_t group_id;
 
 /* The permissions to which the files will be set.  The umask has
    no effect. */
-static mode_t mode = 0755;
+static mode_t mode = S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;
 
 /* If nonzero, strip executable files after copying them. */
 static int strip_files;
@@ -178,26 +167,20 @@ static int strip_files;
 /* If nonzero, install a directory instead of a regular file. */
 static int dir_arg;
 
-/* If nonzero, display usage information and exit.  */
-static int show_help;
-
-/* If nonzero, print the version on standard output and exit.  */
-static int show_version;
-
 static struct option const long_options[] =
 {
-  {"strip", no_argument, NULL, 's'},
+  {"backup", optional_argument, NULL, 'b'},
   {"directory", no_argument, NULL, 'd'},
   {"group", required_argument, NULL, 'g'},
   {"mode", required_argument, NULL, 'm'},
   {"owner", required_argument, NULL, 'o'},
   {"preserve-timestamps", no_argument, NULL, 'p'},
-  {"backup", no_argument, NULL, 'b'},
+  {"strip", no_argument, NULL, 's'},
   {"suffix", required_argument, NULL, 'S'},
-  {"version-control", required_argument, NULL, 'V'},
+  {"version-control", required_argument, NULL, 'V'}, /* Deprecated. FIXME. */
   {"verbose", no_argument, NULL, 'v'},
-  {"help", no_argument, &show_help, 1},
-  {"version", no_argument, &show_version, 1},
+  {GETOPT_HELP_OPTION_DECL},
+  {GETOPT_VERSION_OPTION_DECL},
   {NULL, 0, NULL, 0}
 };
 
@@ -205,8 +188,9 @@ static void
 cp_option_init (struct cp_options *x)
 {
   x->copy_as_regular = 1;
-  x->dereference = 1;
-  x->force = 1;
+  x->dereference = DEREF_ALWAYS;
+  x->unlink_dest_before_opening = 1;
+  x->unlink_dest_after_failed_open = 0;
 
   /* If unlink fails, try to proceed anyway.  */
   x->failed_unlink_is_fatal = 0;
@@ -223,12 +207,13 @@ cp_option_init (struct cp_options *x)
   x->recursive = 0;
   x->sparse_mode = SPARSE_AUTO;
   x->symbolic_link = 0;
+  x->backup_type = none;
 
   /* Create destination files initially writable so we can run strip on them.
      Although GNU strip works fine on read-only files, some others
      would fail.  */
   x->set_mode = 1;
-  x->mode = 0600;
+  x->mode = S_IRUSR | S_IWUSR;
 
   x->umask_kill = 0;
   x->update = 0;
@@ -241,9 +226,10 @@ main (int argc, char **argv)
 {
   int optc;
   int errors = 0;
-  char *symbolic_mode = NULL;
+  const char *symbolic_mode = NULL;
   int make_backups = 0;
-  char *version;
+  char *backup_suffix_string;
+  char *version_control_string = NULL;
   int mkdir_and_install = 0;
   struct cp_options x;
   int n_files;
@@ -254,6 +240,8 @@ main (int argc, char **argv)
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
+  atexit (close_stdout);
+
   cp_option_init (&x);
 
   owner_name = NULL;
@@ -262,10 +250,9 @@ main (int argc, char **argv)
   dir_arg = 0;
   umask (0);
 
-  version = getenv ("SIMPLE_BACKUP_SUFFIX");
-  if (version)
-    simple_backup_suffix = version;
-  version = getenv ("VERSION_CONTROL");
+  /* FIXME: consider not calling getenv for SIMPLE_BACKUP_SUFFIX unless
+     we'll actually use backup_suffix_string.  */
+  backup_suffix_string = getenv ("SIMPLE_BACKUP_SUFFIX");
 
   while ((optc = getopt_long (argc, argv, "bcsDdg:m:o:pvV:S:", long_options,
 			      NULL)) != -1)
@@ -274,8 +261,18 @@ main (int argc, char **argv)
 	{
 	case 0:
 	  break;
+
+	case 'V':  /* FIXME: this is deprecated.  Remove it in 2001.  */
+	  error (0, 0,
+		 _("warning: --version-control (-V) is obsolete;  support for\
+ it\nwill be removed in some future release.  Use --backup=%s instead."
+		   ), optarg);
+	  /* Fall through.  */
+
 	case 'b':
 	  make_backups = 1;
+	  if (optarg)
+	    version_control_string = optarg;
 	  break;
 	case 'c':
 	  break;
@@ -304,32 +301,23 @@ main (int argc, char **argv)
 	  x.preserve_timestamps = 1;
 	  break;
 	case 'S':
-	  simple_backup_suffix = optarg;
-	  break;
-        case 'V':
-	  version = optarg;
+	  make_backups = 1;
+	  backup_suffix_string = optarg;
 	  break;
+	case_GETOPT_HELP_CHAR;
+	case_GETOPT_VERSION_CHAR (PROGRAM_NAME, AUTHORS);
 	default:
 	  usage (1);
 	}
     }
 
-  if (show_version)
-    {
-      printf ("install (%s) %s\n", GNU_PACKAGE, VERSION);
-      close_stdout ();
-      exit (0);
-    }
-
-  if (show_help)
-    usage (0);
-
   /* Check for invalid combinations of arguments. */
   if (dir_arg && strip_files)
     error (1, 0,
 	   _("the strip option may not be used when installing a directory"));
 
-  x.backup_type = (make_backups ? get_version (version) : none);
+  if (make_backups)
+    x.backup_type = xget_version ("backup type", version_control_string);
 
   n_files = argc - optind;
   file = argv + optind;
@@ -344,9 +332,9 @@ main (int argc, char **argv)
     {
       struct mode_change *change = mode_compile (symbolic_mode, 0);
       if (change == MODE_INVALID)
-	error (1, 0, _("invalid mode `%s'"), symbolic_mode);
+	error (1, 0, _("invalid mode %s"), quote (symbolic_mode));
       else if (change == MODE_MEMORY_EXHAUSTED)
-	error (1, 0, _("virtual memory exhausted"));
+	xalloc_die ();
       mode = mode_adjust (0, change);
     }
 
@@ -359,7 +347,7 @@ main (int argc, char **argv)
 	{
 	  errors |=
 	    make_path (file[i], mode, mode, owner_id, group_id, 0,
-		       (x.verbose ? "creating directory `%s'" : NULL));
+		       (x.verbose ? _("creating directory %s") : NULL));
 	}
     }
   else
@@ -384,9 +372,9 @@ main (int argc, char **argv)
 	  if (!isdir (dest))
 	    {
 	      error (0, 0,
-		     _("installing multiple files, but last argument (%s) \
+		     _("installing multiple files, but last argument, %s \
 is not a directory"),
-		     dest);
+		     quote (dest));
 	      usage (1);
 	    }
 	  for (i = 0; i < n_files - 1; i++)
@@ -396,8 +384,6 @@ is not a directory"),
 	}
     }
 
-  if (x.verbose)
-    close_stdout ();
   exit (errors);
 }
 
@@ -409,9 +395,9 @@ install_file_to_path (const char *from, const char *to,
 		      const struct cp_options *x)
 {
   char *dest_dir;
-  int fail;
+  int fail = 0;
 
-  dest_dir = dirname (to);
+  dest_dir = dir_name (to);
 
   /* check to make sure this is a path (not install a b ) */
   if (!STREQ (dest_dir, ".")
@@ -421,17 +407,13 @@ install_file_to_path (const char *from, const char *to,
 	 owner, group, and permissions for parent directories.  Remember
 	 that this option is intended mainly to help installers when the
 	 distribution doesn't provide proper install rules.  */
-#define DIR_MODE 0755
+#define DIR_MODE (S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH)
       fail = make_path (dest_dir, DIR_MODE, DIR_MODE, owner_id, group_id, 0,
-			(x->verbose ? _("creating directory `%s'") : NULL));
+			(x->verbose ? _("creating directory %s") : NULL));
+    }
 
-      if (fail == 0)
-	fail = install_file_in_dir (from, dest_dir, x);
-    }
-  else
-    {
-      fail = install_file_in_file (from, to, x);
-    }
+  if (fail == 0)
+    fail = install_file_in_file (from, to, x);
 
   free (dest_dir);
 
@@ -465,7 +447,7 @@ static int
 install_file_in_dir (const char *from, const char *to_dir,
 		     const struct cp_options *x)
 {
-  char *from_base;
+  const char *from_base;
   char *to;
   int ret;
 
@@ -491,7 +473,7 @@ copy_file (const char *from, const char *to, const struct cp_options *x)
      and that sendmail's installation process relies on the behavior.  */
   if (isdir (from))
     {
-      error (0, 0, _("`%s' is a directory"), from);
+      error (0, 0, _("%s is a directory"), quote (from));
       return 1;
     }
 
@@ -526,15 +508,18 @@ change_attributes (const char *path)
       && errno != EPERM
 #endif
       )
-    err = errno;
-  if (chmod (path, mode))
-    err = errno;
-  if (err)
     {
-      error (0, err, "%s", path);
-      return 1;
+      error (0, errno, "cannot change ownership of %s", quote (path));
+      err = 1;
     }
-  return 0;
+
+  if (!err && chmod (path, mode))
+    {
+      error (0, errno, "cannot change permissions of %s", quote (path));
+      err = 1;
+    }
+
+  return err;
 }
 
 /* Set the timestamps of file TO to match those of file FROM.
@@ -548,7 +533,7 @@ change_timestamps (const char *from, const char *to)
 
   if (stat (from, &stb))
     {
-      error (0, errno, "%s", from);
+      error (0, errno, _("cannot obtain time stamps for %s"), quote (from));
       return 1;
     }
 
@@ -560,7 +545,7 @@ change_timestamps (const char *from, const char *to)
   utb.modtime = stb.st_mtime;
   if (utime (to, &utb))
     {
-      error (0, errno, "%s", to);
+      error (0, errno, _("cannot set time stamps for %s"), quote (to));
       return 1;
     }
   return 0;
@@ -575,13 +560,13 @@ change_timestamps (const char *from, const char *to)
 static void
 strip (const char *path)
 {
-  int pid, status;
+  int status;
+  pid_t pid = fork ();
 
-  pid = fork ();
   switch (pid)
     {
     case -1:
-      error (1, errno, _("fork system call failed"));
+      error (1, errno, _("cannot fork"));
       break;
     case 0:			/* Child. */
       execlp ("strip", "strip", path, NULL);
@@ -591,6 +576,8 @@ strip (const char *path)
       /* Parent process. */
       while (pid != wait (&status))	/* Wait for kid to finish. */
 	/* Do nothing. */ ;
+      if (status)
+	error (1, 0, _("strip failed"));
       break;
     }
 }
@@ -611,7 +598,7 @@ get_ids (void)
 	  long int tmp_long;
 	  if (xstrtol (owner_name, NULL, 0, &tmp_long, NULL) != LONGINT_OK
 	      || tmp_long < 0 || tmp_long > UID_T_MAX)
-	    error (1, 0, _("invalid user `%s'"), owner_name);
+	    error (1, 0, _("invalid user %s"), quote (owner_name));
 	  owner_id = (uid_t) tmp_long;
 	}
       else
@@ -629,7 +616,7 @@ get_ids (void)
 	  long int tmp_long;
 	  if (xstrtol (group_name, NULL, 0, &tmp_long, NULL) != LONGINT_OK
 	      || tmp_long < 0 || tmp_long > GID_T_MAX)
-	    error (1, 0, _("invalid group `%s'"), group_name);
+	    error (1, 0, _("invalid group %s"), quote (group_name));
 	  group_id = (gid_t) tmp_long;
 	}
       else
@@ -640,7 +627,7 @@ get_ids (void)
     group_id = (gid_t) -1;
 }
 
-static void
+void
 usage (int status)
 {
   if (status != 0)
@@ -659,7 +646,8 @@ In the first two formats, copy SOURCE to DEST or multiple SOURCE(s) to\n\
 the existing DIRECTORY, while setting permission modes and owner/group.\n\
 In the third format, create all components of the given DIRECTORY(ies).\n\
 \n\
-  -b, --backup        make backup before removal\n\
+      --backup[=CONTROL] make a backup of each existing destination file\n\
+  -b                  like --backup but does not accept an argument\n\
   -c                  (ignored)\n\
   -d, --directory     treat all arguments as directory names; create all\n\
                         components of the specified directories\n\
@@ -672,22 +660,22 @@ In the third format, create all components of the given DIRECTORY(ies).\n\
                         to corresponding destination files\n\
   -s, --strip         strip symbol tables, only for 1st and 2nd formats\n\
   -S, --suffix=SUFFIX override the usual backup suffix\n\
-      --verbose       print the name of each directory as it is created\n\
-  -V, --version-control=WORD   override the usual version control\n\
+  -v, --verbose       print the name of each directory as it is created\n\
       --help          display this help and exit\n\
       --version       output version information and exit\n\
 \n\
 "));
       printf (_("\
-The backup suffix is ~, unless set with SIMPLE_BACKUP_SUFFIX.  The\n\
-version control may be set with VERSION_CONTROL, values are:\n\
+The backup suffix is `~', unless set with --suffix or SIMPLE_BACKUP_SUFFIX.\n\
+The version control method may be selected via the --backup option or through\n\
+the VERSION_CONTROL environment variable.  Here are the values:\n\
 \n\
-  t, numbered     make numbered backups\n\
-  nil, existing   numbered if numbered backups exist, simple otherwise\n\
-  never, simple   always make simple backups\n\
+  none, off       never make backups (even if --backup is given)\n\
+  numbered, t     make numbered backups\n\
+  existing, nil   numbered if numbered backups exist, simple otherwise\n\
+  simple, never   always make simple backups\n\
 "));
       puts (_("\nReport bugs to <bug-fileutils@gnu.org>."));
-      close_stdout ();
     }
   exit (status);
 }
