@@ -1,5 +1,5 @@
 /* `ln' program to create links between files.
-   Copyright (C) 86, 89, 90, 91, 95, 96, 97, 1998 Free Software Foundation, Inc.
+   Copyright (C) 86, 89, 90, 91, 1995-2000 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -27,9 +27,22 @@
 #include <getopt.h>
 
 #include "system.h"
+#include "same.h"
 #include "backupfile.h"
-#include "closeout.h"
 #include "error.h"
+#include "quote.h"
+
+/* The official name of this program (e.g., no `g' prefix).  */
+#define PROGRAM_NAME "ln"
+
+#define AUTHORS "Mike Parker and David MacKenzie"
+
+/* For long options that have no equivalent short option, use a
+   non-character as a pseudo short option, starting with CHAR_MAX + 1.  */
+enum
+{
+  TARGET_DIRECTORY_OPTION = CHAR_MAX + 1
+};
 
 int link ();			/* Some systems don't declare this anywhere. */
 
@@ -37,13 +50,27 @@ int link ();			/* Some systems don't declare this anywhere. */
 int symlink ();
 #endif
 
+/* In being careful not even to try to make hard links to directories,
+   we have to know whether link(2) follows symlinks.  If it does, then
+   we have to *stat* the `source' to see if the resulting link would be
+   to a directory.  Otherwise, we have to use *lstat* so that we allow
+   users to make hard links to symlinks-that-point-to-directories.  */
+
+#if LINK_FOLLOWS_SYMLINKS
+# define STAT_LIKE_LINK(File, Stat_buf) \
+  stat (File, Stat_buf)
+#else
+# define STAT_LIKE_LINK(File, Stat_buf) \
+  lstat (File, Stat_buf)
+#endif
+
 /* Construct a string NEW_DEST by concatenating DEST, a slash, and
    basename(SOURCE) in alloca'd memory.  Don't modify DEST or SOURCE.  */
 
 #define PATH_BASENAME_CONCAT(new_dest, dest, source)			\
     do									\
       {									\
-	char *source_base;						\
+	const char *source_base;					\
 	char *tmp_source;						\
 									\
 	tmp_source = (char *) alloca (strlen ((source)) + 1);		\
@@ -57,9 +84,6 @@ int symlink ();
       }									\
     while (0)
 
-char *base_name ();
-char *dirname ();
-enum backup_type get_version ();
 int isdir ();
 int yesno ();
 void strip_trailing_slashes ();
@@ -77,10 +101,6 @@ static int (*linkfunc) ();
 /* If nonzero, make symbolic links; otherwise, make hard links.  */
 static int symbolic_link;
 
-/* A string describing type of link to make.  For use in verbose
-   diagnostics and in error messages.  */
-static const char *link_type_string;
-
 /* If nonzero, ask the user before removing existing files.  */
 static int interactive;
 
@@ -99,63 +119,23 @@ static int hard_dir_link;
    symlink-to-dir before creating the new link.  */
 static int dereference_dest_dir_symlinks = 1;
 
-/* If nonzero, display usage information and exit.  */
-static int show_help;
-
-/* If nonzero, print the version on standard output and exit.  */
-static int show_version;
-
 static struct option const long_options[] =
 {
-  {"backup", no_argument, NULL, 'b'},
+  {"backup", optional_argument, NULL, 'b'},
   {"directory", no_argument, NULL, 'F'},
   {"no-dereference", no_argument, NULL, 'n'},
   {"force", no_argument, NULL, 'f'},
   {"interactive", no_argument, NULL, 'i'},
   {"suffix", required_argument, NULL, 'S'},
+  {"target-directory", required_argument, NULL, TARGET_DIRECTORY_OPTION},
   {"symbolic", no_argument, NULL, 's'},
   {"verbose", no_argument, NULL, 'v'},
-  {"version-control", required_argument, NULL, 'V'},
-  {"help", no_argument, &show_help, 1},
-  {"version", no_argument, &show_version, 1},
+  {"version-control", required_argument, NULL, 'V'}, /* Deprecated. FIXME. */
+  {GETOPT_HELP_OPTION_DECL},
+  {GETOPT_VERSION_OPTION_DECL},
   {NULL, 0, NULL, 0}
 };
 
-/* Return nonzero if SOURCE and DEST point to the same name in the same
-   directory.  */
-
-static int
-same_name (const char *source, const char *dest)
-{
-  struct stat source_dir_stats;
-  struct stat dest_dir_stats;
-  char *source_dirname, *dest_dirname;
-
-  source_dirname = dirname (source);
-  dest_dirname = dirname (dest);
-  if (source_dirname == NULL || dest_dirname == NULL)
-    error (1, 0, _("virtual memory exhausted"));
-
-  if (stat (source_dirname, &source_dir_stats))
-    {
-      /* Shouldn't happen.  */
-      error (1, errno, "%s", source_dirname);
-    }
-
-  if (stat (dest_dirname, &dest_dir_stats))
-    {
-      /* Shouldn't happen.  */
-      error (1, errno, "%s", dest_dirname);
-    }
-
-  free (source_dirname);
-  free (dest_dirname);
-
-  return (source_dir_stats.st_dev == dest_dir_stats.st_dev
-	  && source_dir_stats.st_ino == dest_dir_stats.st_ino
-	  && STREQ (base_name (source), base_name (dest)));
-}
-
 /* Make a link DEST to the (usually) existing file SOURCE.
    Symbolic links to nonexistent files are allowed.
    If DEST is a directory, put the link to SOURCE in that directory.
@@ -168,6 +148,7 @@ do_link (const char *source, const char *dest)
   struct stat dest_stats;
   char *dest_backup = NULL;
   int lstat_status;
+  int backup_succeeded = 0;
 
   /* Use stat here instead of lstat.
      On SVR4, link does not follow symlinks, so this check disallows
@@ -175,14 +156,23 @@ do_link (const char *source, const char *dest)
      On other systems, link follows symlinks, so this check is right.  */
   if (!symbolic_link)
     {
-      if (stat (source, &source_stats) != 0)
+      if (STAT_LIKE_LINK (source, &source_stats) != 0)
 	{
-	  error (0, errno, "%s", source);
+	  error (0, errno, _("accessing %s"), quote (source));
 	  return 1;
 	}
+
+      if (S_ISLNK (source_stats.st_mode))
+	{
+	  error (0, 0, _("%s: warning: making a hard link to a symbolic link\
+ is not portable"),
+		 quote (source));
+	}
+
       if (!hard_dir_link && S_ISDIR (source_stats.st_mode))
 	{
-	  error (0, 0, _("%s: hard link not allowed for directory"), source);
+	  error (0, 0, _("%s: hard link not allowed for directory"),
+		 quote (source));
 	  return 1;
 	}
     }
@@ -190,7 +180,7 @@ do_link (const char *source, const char *dest)
   lstat_status = lstat (dest, &dest_stats);
   if (lstat_status != 0 && errno != ENOENT)
     {
-      error (0, errno, "%s", dest);
+      error (0, errno, _("accessing %s"), quote (dest));
       return 1;
     }
 
@@ -217,7 +207,7 @@ do_link (const char *source, const char *dest)
       lstat_status = lstat (dest, &dest_stats);
       if (lstat_status != 0 && errno != ENOENT)
 	{
-	  error (0, errno, "%s", dest);
+	  error (0, errno, _("accessing %s"), quote (dest));
 	  return 1;
 	}
     }
@@ -235,7 +225,7 @@ do_link (const char *source, const char *dest)
 	 beyond this point, because the error message you'd get is
 	 misleading.  */
       && (backup_type == none || !symlink)
-      && (!symlink || stat (source, &source_stats) == 0)
+      && (!symbolic_link || stat (source, &source_stats) == 0)
       && source_stats.st_dev == dest_stats.st_dev
       && source_stats.st_ino == dest_stats.st_ino
       /* The following detects whether removing DEST will also remove
@@ -244,7 +234,8 @@ do_link (const char *source, const char *dest)
 	 name in the same directory.  */
       && (source_stats.st_nlink == 1 || same_name (source, dest)))
     {
-      error (0, 0, _("`%s' and `%s' are the same file"), source, dest);
+      error (0, 0, _("%s and %s are the same file"),
+	     quote_n (0, source), quote_n (1, dest));
       return 1;
     }
 
@@ -252,18 +243,18 @@ do_link (const char *source, const char *dest)
     {
       if (S_ISDIR (dest_stats.st_mode))
 	{
-	  error (0, 0, _("%s: cannot overwrite directory"), dest);
+	  error (0, 0, _("%s: cannot overwrite directory"), quote (dest));
 	  return 1;
 	}
       if (interactive)
 	{
-	  fprintf (stderr, _("%s: replace `%s'? "), program_name, dest);
+	  fprintf (stderr, _("%s: replace %s? "), program_name, quote (dest));
 	  if (!yesno ())
 	    return 0;
 	}
       else if (!remove_existing_files && backup_type == none)
 	{
-	  error (0, 0, _("%s: File exists"), dest);
+	  error (0, 0, _("%s: File exists"), quote (dest));
 	  return 1;
 	}
 
@@ -271,7 +262,7 @@ do_link (const char *source, const char *dest)
 	{
 	  char *tmp_backup = find_backup_file_name (dest, backup_type);
 	  if (tmp_backup == NULL)
-	    error (1, 0, _("virtual memory exhausted"));
+	    xalloc_die ();
 	  dest_backup = (char *) alloca (strlen (tmp_backup) + 1);
 	  strcpy (dest_backup, tmp_backup);
 	  free (tmp_backup);
@@ -279,45 +270,65 @@ do_link (const char *source, const char *dest)
 	    {
 	      if (errno != ENOENT)
 		{
-		  error (0, errno, _("cannot backup `%s'"), dest);
+		  error (0, errno, _("cannot backup %s"), quote (dest));
 		  return 1;
 		}
 	      else
 		dest_backup = NULL;
 	    }
+	  else
+	    {
+	      backup_succeeded = 1;
+	    }
 	}
-      else if (unlink (dest) && errno != ENOENT)
+
+      /* Try to unlink DEST even if we may have renamed it.  In some unusual
+	 cases (when DEST and DEST_BACKUP are hard-links that refer to the
+	 same file), rename succeeds and DEST remains.  If we didn't remove
+	 DEST in that case, the subsequent LINKFUNC call would fail.  */
+      if (unlink (dest) && errno != ENOENT)
 	{
-	  error (0, errno, _("cannot remove `%s'"), dest);
+	  error (0, errno, _("cannot remove %s"), quote (dest));
 	  return 1;
 	}
     }
   else if (errno != ENOENT)
     {
-      error (0, errno, "%s", dest);
+      error (0, errno, _("accessing %s"), quote (dest));
       return 1;
     }
 
   if (verbose)
-    printf (_("create %s %s to %s\n"), link_type_string, dest, source);
+    {
+      printf ((symbolic_link
+	       ? _("create symbolic link %s to %s")
+	       : _("create hard link %s to %s")),
+	      quote_n (0, dest), quote_n (1, source));
+      if (backup_succeeded)
+	printf (_(" (backup: %s)"), quote (dest_backup));
+      putchar ('\n');
+    }
 
   if ((*linkfunc) (source, dest) == 0)
     {
       return 0;
     }
 
-  error (0, errno, _("cannot create %s `%s' to `%s'"), link_type_string,
-	 dest, source);
+  error (0, errno,
+	 (symbolic_link
+	  ? _("creating symbolic link %s to %s")
+	  : _("creating hard link %s to %s")),
+	 quote_n (0, dest), quote_n (1, source));
 
   if (dest_backup)
     {
       if (rename (dest_backup, dest))
-	error (0, errno, _("cannot un-backup `%s'"), dest);
+	error (0, errno, _("cannot un-backup %s"), quote (dest));
     }
   return 1;
 }
 
-static void
+void
 usage (int status)
 {
   if (status != 0)
@@ -328,15 +339,19 @@ usage (int status)
       printf (_("\
 Usage: %s [OPTION]... TARGET [LINK_NAME]\n\
   or:  %s [OPTION]... TARGET... DIRECTORY\n\
+  or:  %s [OPTION]... --target-directory=DIRECTORY TARGET...\n\
 "),
-	      program_name, program_name);
+	      program_name, program_name, program_name);
       printf (_("\
-Create a link to the specified TARGET with optional LINK_NAME.  If there is\n\
-more than one TARGET, the last argument must be a directory;  create links\n\
-in DIRECTORY to each TARGET.  Create hard links by default, symbolic links\n\
-with --symbolic.  When creating hard links, each TARGET must exist.\n\
+Create a link to the specified TARGET with optional LINK_NAME.\n\
+If LINK_NAME is omitted, a link with the same basename as the TARGET is\n\
+created in the current directory.  When using the second form with more\n\
+than one TARGET, the last argument must be a directory;  create links\n\
+in DIRECTORY to each TARGET.  Create hard links by default, symbolic\n\
+links with --symbolic.  When creating hard links, each TARGET must exist.\n\
 \n\
-  -b, --backup                make a backup of each existing destination file\n\
+      --backup[=CONTROL]      make a backup of each existing destination file\n\
+  -b                          like --backup but does not accept an argument\n\
   -d, -F, --directory         hard link directories (super-user only)\n\
   -f, --force                 remove existing destination files\n\
   -n, --no-dereference        treat destination that is a symlink to a\n\
@@ -344,22 +359,24 @@ with --symbolic.  When creating hard links, each TARGET must exist.\n\
   -i, --interactive           prompt whether to remove destinations\n\
   -s, --symbolic              make symbolic links instead of hard links\n\
   -S, --suffix=SUFFIX         override the usual backup suffix\n\
+      --target-directory=DIRECTORY  specify the DIRECTORY in which to create\n\
+                                the links\n\
   -v, --verbose               print name of each file before linking\n\
-  -V, --version-control=WORD  override the usual version control\n\
       --help                  display this help and exit\n\
       --version               output version information and exit\n\
 \n\
 "));
       printf (_("\
-The backup suffix is ~, unless set with SIMPLE_BACKUP_SUFFIX.  The\n\
-version control may be set with VERSION_CONTROL, values are:\n\
+The backup suffix is `~', unless set with --suffix or SIMPLE_BACKUP_SUFFIX.\n\
+The version control method may be selected via the --backup option or through\n\
+the VERSION_CONTROL environment variable.  Here are the values:\n\
 \n\
-  t, numbered     make numbered backups\n\
-  nil, existing   numbered if numbered backups exist, simple otherwise\n\
-  never, simple   always make simple backups\n\
+  none, off       never make backups (even if --backup is given)\n\
+  numbered, t     make numbered backups\n\
+  existing, nil   numbered if numbered backups exist, simple otherwise\n\
+  simple, never   always make simple backups\n\
 "));
       puts (_("\nReport bugs to <bug-fileutils@gnu.org>."));
-      close_stdout ();
     }
   exit (status);
 }
@@ -370,17 +387,24 @@ main (int argc, char **argv)
   int c;
   int errors;
   int make_backups = 0;
-  char *version;
+  char *backup_suffix_string;
+  char *version_control_string = NULL;
+  char *target_directory = NULL;
+  int target_directory_specified;
+  unsigned int n_files;
+  char **file;
+  int dest_is_dir;
 
   program_name = argv[0];
   setlocale (LC_ALL, "");
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
-  version = getenv ("SIMPLE_BACKUP_SUFFIX");
-  if (version)
-    simple_backup_suffix = version;
-  version = getenv ("VERSION_CONTROL");
+  atexit (close_stdout);
+
+  /* FIXME: consider not calling getenv for SIMPLE_BACKUP_SUFFIX unless
+     we'll actually use backup_suffix_string.  */
+  backup_suffix_string = getenv ("SIMPLE_BACKUP_SUFFIX");
 
   symbolic_link = remove_existing_files = interactive = verbose
     = hard_dir_link = 0;
@@ -393,8 +417,18 @@ main (int argc, char **argv)
 	{
 	case 0:			/* Long-named option. */
  	  break;
+
+	case 'V':  /* FIXME: this is deprecated.  Remove it in 2001.  */
+	  error (0, 0,
+		 _("warning: --version-control (-V) is obsolete;  support for\
+ it\nwill be removed in some future release.  Use --backup=%s instead."
+		   ), optarg);
+	  /* Fall through.  */
+
 	case 'b':
 	  make_backups = 1;
+	  if (optarg)
+	    version_control_string = optarg;
 	  break;
 	case 'd':
 	case 'F':
@@ -418,65 +452,91 @@ main (int argc, char **argv)
 	  error (1, 0, _("symbolic links are not supported on this system"));
 #endif
 	  break;
+	case TARGET_DIRECTORY_OPTION:
+	  target_directory = optarg;
+	  break;
 	case 'v':
 	  verbose = 1;
 	  break;
 	case 'S':
-	  simple_backup_suffix = optarg;
-	  break;
-	case 'V':
-	  version = optarg;
+	  make_backups = 1;
+	  backup_suffix_string = optarg;
 	  break;
+	case_GETOPT_HELP_CHAR;
+	case_GETOPT_VERSION_CHAR (PROGRAM_NAME, AUTHORS);
 	default:
 	  usage (1);
 	  break;
 	}
     }
 
-  if (show_version)
-    {
-      printf ("ln (%s) %s\n", GNU_PACKAGE, VERSION);
-      close_stdout ();
-      exit (0);
-    }
+  n_files = argc - optind;
+  file = argv + optind;
 
-  if (show_help)
-    usage (0);
-
-  if (optind == argc)
+  if (n_files == 0)
     {
       error (0, 0, _("missing file argument"));
       usage (1);
     }
 
-  backup_type = (make_backups ? get_version (version) : none);
+  target_directory_specified = (target_directory != NULL);
+  if (!target_directory)
+    target_directory = file[n_files - 1];
 
-#ifdef S_ISLNK
-  if (symbolic_link)
+  /* If there's only one file argument, then pretend `.' was given
+     as the second argument.  */
+  if (n_files == 1)
     {
-      linkfunc = symlink;
-      link_type_string = _("symbolic link");
+      static char *dummy[2];
+      dummy[0] = file[0];
+      dummy[1] = ".";
+      file = dummy;
+      n_files = 2;
+      dest_is_dir = 1;
     }
   else
     {
-      linkfunc = link;
-      link_type_string = _("hard link");
+      dest_is_dir = isdir (target_directory);
     }
-#else
-  link_type_string = _("link");
-#endif
 
-  if (optind == argc - 1)
-    errors = do_link (argv[optind], ".");
-  else if (optind == argc - 2)
+  if (symbolic_link)
+    linkfunc = symlink;
+  else
+    linkfunc = link;
+
+  if (target_directory_specified && !dest_is_dir)
+    {
+      error (0, 0, _("specified target directory, %s is not a directory"),
+	     quote (target_directory));
+      usage (1);
+    }
+
+  backup_type = (make_backups
+		 ? xget_version (_("backup type"), version_control_string)
+		 : none);
+
+  if (target_directory_specified || n_files > 2)
+    {
+      int i;
+      unsigned int last_file_idx = (target_directory_specified
+				    ? n_files - 1
+				    : n_files - 2);
+
+      if (!target_directory_specified && !dest_is_dir)
+	error (1, 0,
+	   _("when making multiple links, last argument must be a directory"));
+      for (i = 0; i <= last_file_idx; ++i)
+	errors += do_link (file[i], target_directory);
+    }
+  else
     {
       struct stat source_stats;
-      char *source;
+      const char *source;
       char *dest;
       char *new_dest;
 
-      source = argv[optind];
-      dest = argv[optind + 1];
+      source = file[0];
+      dest = file[1];
 
       /* When the destination is specified with a trailing slash and the
 	 source exists but is not a directory, convert the user's command
@@ -495,18 +555,6 @@ main (int argc, char **argv)
 
       errors = do_link (source, new_dest);
     }
-  else
-    {
-      char *to;
 
-      to = argv[argc - 1];
-      if (!isdir (to))
-	error (1, 0, _("when making multiple links, last argument must be a directory"));
-      for (; optind < argc - 1; ++optind)
-	errors += do_link (argv[optind], to);
-    }
-
-  if (verbose)
-    close_stdout ();
   exit (errors != 0);
 }
