@@ -1,5 +1,5 @@
 /* mkdir -- make directories
-   Copyright (C) 90, 95, 96, 97, 1998 Free Software Foundation, Inc.
+   Copyright (C) 90, 1995-2000 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -23,34 +23,33 @@
 #include <sys/types.h>
 
 #include "system.h"
-#include "modechange.h"
-#include "makepath.h"
-#include "closeout.h"
 #include "error.h"
+#include "makepath.h"
+#include "modechange.h"
+#include "quote.h"
+
+/* The official name of this program (e.g., no `g' prefix).  */
+#define PROGRAM_NAME "mkdir"
+
+#define AUTHORS "David MacKenzie"
 
 /* The name this program was run with. */
 char *program_name;
 
 /* If nonzero, ensure that all parents of the specified directory exist.  */
-static int path_mode;
-
-/* If nonzero, display usage information and exit.  */
-static int show_help;
-
-/* If nonzero, print the version on standard output and exit.  */
-static int show_version;
+static int create_parents;
 
 static struct option const longopts[] =
 {
   {"mode", required_argument, NULL, 'm'},
   {"parents", no_argument, NULL, 'p'},
-  {"help", no_argument, &show_help, 1},
-  {"verbose", no_argument, NULL, 2},
-  {"version", no_argument, &show_version, 1},
+  {"verbose", no_argument, NULL, 'v'},
+  {GETOPT_HELP_OPTION_DECL},
+  {GETOPT_VERSION_OPTION_DECL},
   {NULL, 0, NULL, 0}
 };
 
-static void
+void
 usage (int status)
 {
   if (status != 0)
@@ -64,12 +63,11 @@ Create the DIRECTORY(ies), if they do not already exist.\n\
 \n\
   -m, --mode=MODE   set permission mode (as in chmod), not rwxrwxrwx - umask\n\
   -p, --parents     no error if existing, make parent directories as needed\n\
-      --verbose     print a message for each created directory\n\
+  -v, --verbose     print a message for each created directory\n\
       --help        display this help and exit\n\
       --version     output version information and exit\n\
 "));
       puts (_("\nReport bugs to <bug-fileutils@gnu.org>."));
-      close_stdout ();
     }
   exit (status);
 }
@@ -77,9 +75,9 @@ Create the DIRECTORY(ies), if they do not already exist.\n\
 int
 main (int argc, char **argv)
 {
-  unsigned int newmode;
-  unsigned int parent_mode;
-  char *symbolic_mode = NULL;
+  mode_t newmode;
+  mode_t parent_mode;
+  const char *symbolic_mode = NULL;
   const char *verbose_fmt_string = NULL;
   int errors = 0;
   int optc;
@@ -89,72 +87,80 @@ main (int argc, char **argv)
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
-  path_mode = 0;
+  atexit (close_stdout);
 
-  while ((optc = getopt_long (argc, argv, "pm:", longopts, NULL)) != -1)
+  create_parents = 0;
+
+  while ((optc = getopt_long (argc, argv, "pm:v", longopts, NULL)) != -1)
     {
       switch (optc)
 	{
 	case 0:			/* Long option. */
 	  break;
 	case 'p':
-	  path_mode = 1;
+	  create_parents = 1;
 	  break;
 	case 'm':
 	  symbolic_mode = optarg;
 	  break;
-	case 2: /* --verbose  */
-	  verbose_fmt_string = _("created directory `%s'");
+	case 'v': /* --verbose  */
+	  verbose_fmt_string = _("created directory %s");
 	  break;
+	case_GETOPT_HELP_CHAR;
+	case_GETOPT_VERSION_CHAR (PROGRAM_NAME, AUTHORS);
 	default:
 	  usage (1);
 	}
     }
 
-  if (show_version)
-    {
-      printf ("mkdir (%s) %s\n", GNU_PACKAGE, VERSION);
-      close_stdout ();
-      exit (0);
-    }
-
-  if (show_help)
-    usage (0);
-
   if (optind == argc)
     {
       error (0, 0, _("too few arguments"));
       usage (1);
     }
 
-  newmode = 0777 & ~umask (0);
-  parent_mode = newmode | 0300;	/* u+wx */
+  newmode = S_IRWXUGO & ~ umask (0);
+  parent_mode = S_IWUSR | S_IXUSR | newmode;
   if (symbolic_mode)
     {
       struct mode_change *change = mode_compile (symbolic_mode, 0);
       if (change == MODE_INVALID)
-	error (1, 0, _("invalid mode `%s'"), symbolic_mode);
+	error (1, 0, _("invalid mode %s"), quote (symbolic_mode));
       else if (change == MODE_MEMORY_EXHAUSTED)
-	error (1, 0, _("virtual memory exhausted"));
+	xalloc_die ();
       newmode = mode_adjust (newmode, change);
     }
 
   for (; optind < argc; ++optind)
     {
-      if (path_mode)
+      int fail = 0;
+      if (create_parents)
 	{
-	  errors |= make_path (argv[optind], newmode, parent_mode,
-			       -1, -1, 1, verbose_fmt_string);
+	  fail = make_path (argv[optind], newmode, parent_mode,
+			    -1, -1, 1, verbose_fmt_string);
 	}
-      else if (mkdir (argv[optind], newmode))
+      else
 	{
-	  error (0, errno, _("cannot make directory `%s'"), argv[optind]);
-	  errors = 1;
-	}
-      else if (verbose_fmt_string)
-	{
-	  error (0, 0, verbose_fmt_string, argv[optind]);
+	  fail = mkdir (argv[optind], newmode);
+	  if (fail)
+	    error (0, errno, _("cannot create directory %s"),
+		   quote (argv[optind]));
+	  else if (verbose_fmt_string)
+	    error (0, 0, verbose_fmt_string, quote (argv[optind]));
+
+	  /* mkdir(2) is required to honor only the file permission bits.
+	     In particular, it needn't do anything about `special' bits,
+	     so if any were set in newmode, apply them with chmod.  */
+	  if (fail == 0 && (newmode & ~S_IRWXUGO))
+	    {
+	      fail = chmod (argv[optind], newmode);
+	      if (fail)
+		error (0, errno, _("cannot set permissions of directory %s"),
+		       quote (argv[optind]));
+	    }
 	}
+
+      errors |= fail;
     }
 
   exit (errors);
