@@ -1,5 +1,5 @@
 /* mv -- move or rename files
-   Copyright (C) 86, 89, 90, 91, 95, 96, 97, 1998 Free Software Foundation, Inc.
+   Copyright (C) 86, 89, 90, 91, 1995-2000 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -15,27 +15,7 @@
    along with this program; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
 
-/* Options:
-   -f, --force		Assume a 'y' answer to all questions it would
-			normally ask, and not ask the questions.
-
-   -i, --interactive	Require confirmation from the user before
-			performing any move that would destroy an
-			existing file.
-
-   -u, --update		Do not move a nondirectory that has an
-			existing destination with the same or newer
-			modification time.
-
-   -v, --verbose		List the name of each file as it is moved, and
-			the name it is moved to.
-
-   -b, --backup
-   -S, --suffix
-   -V, --version-control
-			Backup file creation.
-
-   Written by Mike Parker and David MacKenzie */
+/* Written by Mike Parker, David MacKenzie, and Jim Meyering */
 
 #ifdef _AIX
  #pragma alloca
@@ -48,13 +28,18 @@
 #include <assert.h>
 
 #include "system.h"
-#include "path-concat.h"
 #include "backupfile.h"
-#include "closeout.h"
-#include "cp-hash.h"
 #include "copy.h"
-#include "remove.h"
+#include "cp-hash.h"
 #include "error.h"
+#include "path-concat.h"
+#include "quote.h"
+#include "remove.h"
+
+/* The official name of this program (e.g., no `g' prefix).  */
+#define PROGRAM_NAME "mv"
+
+#define AUTHORS "Mike Parker, David MacKenzie, and Jim Meyering"
 
 /* Initial number of entries in each hash table entry's table of inodes.  */
 #define INITIAL_HASH_MODULE 100
@@ -62,10 +47,16 @@
 /* Initial number of entries in the inode hash table.  */
 #define INITIAL_ENTRY_TAB_SIZE 70
 
-char *base_name ();
+/* For long options that have no equivalent short option, use a
+   non-character as a pseudo short option, starting with CHAR_MAX + 1.  */
+enum
+{
+  TARGET_DIRECTORY_OPTION = CHAR_MAX + 1,
+  STRIP_TRAILING_SLASHES_OPTION
+};
+
 int euidaccess ();
 int full_write ();
-enum backup_type get_version ();
 int isdir ();
 int lstat ();
 int yesno ();
@@ -73,26 +64,22 @@ int yesno ();
 /* The name this program was run with. */
 char *program_name;
 
-/* If nonzero, stdin is a tty. */
-static int stdin_tty;
-
-/* If nonzero, display usage information and exit.  */
-static int show_help;
-
-/* If nonzero, print the version on standard output and exit.  */
-static int show_version;
+/* Remove any trailing slashes from each SOURCE argument.  */
+static int remove_trailing_slashes;
 
 static struct option const long_options[] =
 {
-  {"backup", no_argument, NULL, 'b'},
+  {"backup", optional_argument, NULL, 'b'},
   {"force", no_argument, NULL, 'f'},
   {"interactive", no_argument, NULL, 'i'},
+  {"strip-trailing-slash", no_argument, NULL, STRIP_TRAILING_SLASHES_OPTION},
   {"suffix", required_argument, NULL, 'S'},
+  {"target-directory", required_argument, NULL, TARGET_DIRECTORY_OPTION},
   {"update", no_argument, NULL, 'u'},
   {"verbose", no_argument, NULL, 'v'},
   {"version-control", required_argument, NULL, 'V'},
-  {"help", no_argument, &show_help, 1},
-  {"version", no_argument, &show_version, 1},
+  {GETOPT_HELP_OPTION_DECL},
+  {GETOPT_VERSION_OPTION_DECL},
   {NULL, 0, NULL, 0}
 };
 
@@ -118,8 +105,9 @@ static void
 cp_option_init (struct cp_options *x)
 {
   x->copy_as_regular = 0;  /* FIXME: maybe make this an option */
-  x->dereference = 0;
-  x->force = 0;
+  x->dereference = DEREF_NEVER;
+  x->unlink_dest_before_opening = 0;
+  x->unlink_dest_after_failed_open = 0;
   x->failed_unlink_is_fatal = 1;
   x->hard_link = 0;
   x->interactive = 0;
@@ -140,7 +128,7 @@ cp_option_init (struct cp_options *x)
      when using chmod.  The creation mask is set to be liberal, so
      that created directories can be written, even if it would not
      have been allowed with the mask this process was started with.  */
-  x->umask_kill = 0777777 ^ umask (0);
+  x->umask_kill = ~ umask (0);
 
   x->update = 0;
   x->verbose = 0;
@@ -157,6 +145,18 @@ is_real_dir (const char *path)
   return lstat (path, &stats) == 0 && S_ISDIR (stats.st_mode);
 }
 
+static int
+strip_trailing_slashes_2 (char *path)
+{
+  char *end_p = path + strlen (path) - 1;
+  char *slash = end_p;
+
+  while (slash > path && *slash == '/')
+    *slash-- = '\0';
+
+  return slash < end_p;
+}
+
 /* Move SOURCE onto DEST.  Handles cross-filesystem moves.
    If SOURCE is a directory, DEST must not exist.
    Return 0 if successful, non-zero if an error occurred.  */
@@ -195,10 +195,10 @@ do_move (const char *source, const char *dest, const struct cp_options *x)
 	     copied-into-self directory, DEST (`b/b' in the example),
 	     and failing.  */
 
-	  dir_to_remove = dest;
+	  dir_to_remove = NULL;
 	  error (0, 0,
-		 _("cannot move `%s' to a subdirectory of itself, `%s'"),
-		 source, dest);
+		 _("cannot move %s to a subdirectory of itself, %s"),
+		 quote_n (0, source), quote_n (1, dest));
 	}
       else if (rename_succeeded)
 	{
@@ -245,6 +245,11 @@ do_move (const char *source, const char *dest, const struct cp_options *x)
 	  remove_init ();
 
 	  fspec_init_file (&fs, dir_to_remove);
+
+	  /* Remove any trailing slashes.  This is necessary if we
+	     took the else branch of movefile.  */
+	  strip_trailing_slashes_2 (fs.filename);
+
 	  status = rm (&fs, 1, &rm_options);
 	  assert (VALID_STATUS (status));
 	  if (status == RM_ERROR)
@@ -253,7 +258,7 @@ do_move (const char *source, const char *dest, const struct cp_options *x)
 	  remove_fini ();
 
 	  if (fail)
-	    error (0, errno, _("cannot remove `%s'"), dir_to_remove);
+	    error (0, errno, _("cannot remove %s"), quote (dir_to_remove));
 	}
 
       if (copy_into_self)
@@ -263,29 +268,28 @@ do_move (const char *source, const char *dest, const struct cp_options *x)
   return fail;
 }
 
-static int
-strip_trailing_slashes_2 (char *path)
-{
-  char *end_p = path + strlen (path) - 1;
-  char *slash = end_p;
-
-  while (slash > path && *slash == '/')
-    *slash-- = '\0';
-
-  return slash < end_p;
-}
-
 /* Move file SOURCE onto DEST.  Handles the case when DEST is a directory.
    DEST_IS_DIR must be nonzero when DEST is a directory or a symlink to a
    directory and zero otherwise.
    Return 0 if successful, non-zero if an error occurred.  */
 
 static int
-movefile (char *source, char *dest, int dest_is_dir, const struct cp_options *x)
+movefile (char *source, char *dest, int dest_is_dir,
+	  const struct cp_options *x)
 {
   int dest_had_trailing_slash = strip_trailing_slashes_2 (dest);
   int fail;
 
+  /* This code was introduced to handle the ambiguity in the semantics
+     of mv that is induced by the varying semantics of the rename function.
+     Some systems (e.g., Linux) have a rename function that honors a
+     trailing slash, while others (like Solaris 5,6,7) have a rename
+     function that ignores a trailing slash.  I believe the Linux
+     rename semantics are POSIX and susv2 compliant.  */
+
+  if (remove_trailing_slashes)
+    strip_trailing_slashes_2 (source);
+
   /* In addition to when DEST is a directory, if DEST has a trailing
      slash and neither SOURCE nor DEST is a directory, presume the target
      is DEST/`basename source`.  This converts `mv x y/' to `mv x y/x'.
@@ -306,9 +310,11 @@ movefile (char *source, char *dest, int dest_is_dir, const struct cp_options *x)
       src_basename = base_name (source);
       new_dest = path_concat (dest, src_basename, NULL);
       if (new_dest == NULL)
-	error (1, 0, _("virtual memory exhausted"));
+	xalloc_die ();
       fail = do_move (source, new_dest, x);
-      free (new_dest);
+
+      /* Do not free new_dest.  It may have been squirreled away by
+	 the remember_copied function.  */
     }
   else
     {
@@ -318,7 +324,7 @@ movefile (char *source, char *dest, int dest_is_dir, const struct cp_options *x)
   return fail;
 }
 
-static void
+void
 usage (int status)
 {
   if (status != 0)
@@ -329,32 +335,37 @@ usage (int status)
       printf (_("\
 Usage: %s [OPTION]... SOURCE DEST\n\
   or:  %s [OPTION]... SOURCE... DIRECTORY\n\
+  or:  %s [OPTION]... --target-directory=DIRECTORY SOURCE...\n\
 "),
-	      program_name, program_name);
+	      program_name, program_name, program_name);
       printf (_("\
 Rename SOURCE to DEST, or move SOURCE(s) to DIRECTORY.\n\
 \n\
-  -b, --backup                 make backup before removal\n\
-  -f, --force                  remove existing destinations, never prompt\n\
+      --backup[=CONTROL]       make a backup of each existing destination file\n\
+  -b                           like --backup but does not accept an argument\n\
+  -f, --force                  never prompt before overwriting\n\
   -i, --interactive            prompt before overwrite\n\
+      --strip-trailing-slashes  remove any trailing slashes from each SOURCE\n\
+                                 argument\n\
   -S, --suffix=SUFFIX          override the usual backup suffix\n\
+      --target-directory=DIRECTORY  move all SOURCE arguments into DIRECTORY\n\
   -u, --update                 move only older or brand new non-directories\n\
   -v, --verbose                explain what is being done\n\
-  -V, --version-control=WORD   override the usual version control\n\
       --help                   display this help and exit\n\
       --version                output version information and exit\n\
 \n\
 "));
       printf (_("\
-The backup suffix is ~, unless set with SIMPLE_BACKUP_SUFFIX.  The\n\
-version control may be set with VERSION_CONTROL, values are:\n\
+The backup suffix is `~', unless set with --suffix or SIMPLE_BACKUP_SUFFIX.\n\
+The version control method may be selected via the --backup option or through\n\
+the VERSION_CONTROL environment variable.  Here are the values:\n\
 \n\
-  t, numbered     make numbered backups\n\
-  nil, existing   numbered if numbered backups exist, simple otherwise\n\
-  never, simple   always make simple backups\n\
+  none, off       never make backups (even if --backup is given)\n\
+  numbered, t     make numbered backups\n\
+  existing, nil   numbered if numbered backups exist, simple otherwise\n\
+  simple, never   always make simple backups\n\
 "));
       puts (_("\nReport bugs to <bug-fileutils@gnu.org>."));
-      close_stdout ();
     }
   exit (status);
 }
@@ -366,20 +377,26 @@ main (int argc, char **argv)
   int errors;
   int make_backups = 0;
   int dest_is_dir;
-  char *version;
+  char *backup_suffix_string;
+  char *version_control_string = NULL;
   struct cp_options x;
+  char *target_directory = NULL;
+  int target_directory_specified;
+  unsigned int n_files;
+  char **file;
 
   program_name = argv[0];
   setlocale (LC_ALL, "");
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
+  atexit (close_stdout);
+
   cp_option_init (&x);
 
-  version = getenv ("SIMPLE_BACKUP_SUFFIX");
-  if (version)
-    simple_backup_suffix = version;
-  version = getenv ("VERSION_CONTROL");
+  /* FIXME: consider not calling getenv for SIMPLE_BACKUP_SUFFIX unless
+     we'll actually use backup_suffix_string.  */
+  backup_suffix_string = getenv ("SIMPLE_BACKUP_SUFFIX");
 
   errors = 0;
 
@@ -389,16 +406,30 @@ main (int argc, char **argv)
 	{
 	case 0:
 	  break;
+
+	case 'V':  /* FIXME: this is deprecated.  Remove it in 2001.  */
+	  error (0, 0,
+		 _("warning: --version-control (-V) is obsolete;  support for\
+ it\nwill be removed in some future release.  Use --backup=%s instead."
+		   ), optarg);
+	  /* Fall through.  */
+
 	case 'b':
 	  make_backups = 1;
+	  if (optarg)
+	    version_control_string = optarg;
 	  break;
 	case 'f':
 	  x.interactive = 0;
-	  x.force = 1;
 	  break;
 	case 'i':
 	  x.interactive = 1;
-	  x.force = 0;
+	  break;
+	case STRIP_TRAILING_SLASHES_OPTION:
+	  remove_trailing_slashes = 1;
+	  break;
+	case TARGET_DIRECTORY_OPTION:
+	  target_directory = optarg;
 	  break;
 	case 'u':
 	  x.update = 1;
@@ -407,48 +438,64 @@ main (int argc, char **argv)
 	  x.verbose = 1;
 	  break;
 	case 'S':
-	  simple_backup_suffix = optarg;
-	  break;
-	case 'V':
-	  version = optarg;
+	  make_backups = 1;
+	  backup_suffix_string = optarg;
 	  break;
+	case_GETOPT_HELP_CHAR;
+	case_GETOPT_VERSION_CHAR (PROGRAM_NAME, AUTHORS);
 	default:
 	  usage (1);
 	}
     }
 
-  if (show_version)
-    {
-      printf ("mv (%s) %s\n", GNU_PACKAGE, VERSION);
-      close_stdout ();
-      exit (0);
-    }
+  n_files = argc - optind;
+  file = argv + optind;
 
-  if (show_help)
-    usage (0);
+  target_directory_specified = (target_directory != NULL);
+  if (target_directory == NULL && n_files != 0)
+    target_directory = file[n_files - 1];
 
-  if (argc < optind + 2)
+  dest_is_dir = (n_files > 0 && isdir (target_directory));
+
+  if (n_files == 0 || (n_files == 1 && !target_directory_specified))
     {
-      error (0, 0, "%s", (argc == optind
-			  ? _("missing file arguments")
-			  : _("missing file argument")));
+      error (0, 0, _("missing file argument"));
       usage (1);
     }
 
-  x.backup_type = (make_backups ? get_version (version) : none);
+  if (target_directory_specified)
+    {
+      if (!dest_is_dir)
+	{
+	  error (0, 0, _("specified target, %s is not a directory"),
+		 quote (target_directory));
+	  usage (1);
+	}
+    }
+  else if (n_files > 2 && !dest_is_dir)
+    {
+      error (0, 0,
+	    _("when moving multiple files, last argument must be a directory"));
+      usage (1);
+    }
 
-  stdin_tty = isatty (STDIN_FILENO);
-  dest_is_dir = isdir (argv[argc - 1]);
+  if (backup_suffix_string)
+    simple_backup_suffix = xstrdup (backup_suffix_string);
 
-  if (argc > optind + 2 && !dest_is_dir)
-    error (1, 0,
-	   _("when moving multiple files, last argument must be a directory"));
+  x.backup_type = (make_backups
+		   ? xget_version (_("backup type"),
+				   version_control_string)
+		   : none);
 
-  /* Move each arg but the last onto the last. */
-  for (; optind < argc - 1; ++optind)
-    errors |= movefile (argv[optind], argv[argc - 1], dest_is_dir, &x);
+  /* Move each arg but the last into the target_directory.  */
+  {
+    unsigned int last_file_idx = (target_directory_specified
+				  ? n_files - 1
+				  : n_files - 2);
+    unsigned int i;
+    for (i = 0; i <= last_file_idx; ++i)
+      errors |= movefile (file[i], target_directory, dest_is_dir, &x);
+  }
 
-  if (x.verbose)
-    close_stdout ();
   exit (errors);
 }
