@@ -1,5 +1,5 @@
 /* remove.c -- core functions for removing files and directories
-   Copyright (C) 88, 90, 91, 94, 95, 96, 97, 1998 Free Software Foundation, Inc.
+   Copyright (C) 88, 90, 91, 1994-2000 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -34,10 +34,10 @@ typedef enum {false = 0, true = 1} bool;
 
 #include "save-cwd.h"
 #include "system.h"
-#include "closeout.h"
 #include "error.h"
 #include "obstack.h"
 #include "hash.h"
+#include "quote.h"
 #include "remove.h"
 
 #define obstack_chunk_alloc malloc
@@ -66,29 +66,10 @@ int rpl_lstat PARAMS((const char *, struct stat *));
 # define D_INO(dp) 1
 #endif
 
-#if !defined (S_ISLNK)
+#if !defined S_ISLNK
 # define S_ISLNK(Mode) 0
 #endif
 
-#define DOT_OR_DOTDOT(Basename) \
-  (Basename[0] == '.' && (Basename[1] == '\0' \
-			  || (Basename[1] == '.' && Basename[2] == '\0')))
-
-#if defined strdupa
-# define ASSIGN_STRDUPA(DEST, S)		\
-  do { DEST = strdupa(S); } while (0)
-#else
-# define ASSIGN_STRDUPA(DEST, S)		\
-  do						\
-    {						\
-      const char *s_ = (S);			\
-      size_t len_ = strlen (s_) + 1;		\
-      char *tmp_dest_ = (char *) alloca (len_);	\
-      DEST = memcpy (tmp_dest_, (s_), len_);	\
-    }						\
-  while (0)
-#endif
-
 /* Initial capacity of per-directory hash table of entries that have
    been processed but not been deleted.  */
 #define HT_INITIAL_CAPACITY 13
@@ -97,7 +78,6 @@ int rpl_lstat PARAMS((const char *, struct stat *));
    be resized only for hierarchies more than about 45 levels deep. */
 #define ACTIVE_DIR_INITIAL_CAPACITY 53
 
-char *base_name ();
 int euidaccess ();
 int yesno ();
 
@@ -108,7 +88,8 @@ extern char *program_name;
 /* An entry in the active_dir_map.  */
 struct active_dir_ent
 {
-  ino_t inum;
+  ino_t st_ino;
+  dev_t st_dev;
   unsigned int depth;
 };
 
@@ -162,15 +143,16 @@ print_nth_dir (FILE *stream, unsigned int depth)
       sum += length[i];
     }
 
-  fwrite (dir_name, 1, sum, stream);
+  fwrite (dir_name, 1, sum - 1, stream);
 }
 
 static inline struct active_dir_ent *
-make_active_dir_ent (ino_t inum, unsigned int depth)
+make_active_dir_ent (ino_t inum, dev_t device, unsigned int depth)
 {
   struct active_dir_ent *ent;
   ent = (struct active_dir_ent *) xmalloc (sizeof *ent);
-  ent->inum = inum;
+  ent->st_ino = inum;
+  ent->st_dev = device;
   ent->depth = depth;
   return ent;
 }
@@ -179,7 +161,9 @@ static unsigned int
 hash_active_dir_ent (void const *x, unsigned int table_size)
 {
   struct active_dir_ent const *ade = x;
-  return ade->inum % table_size;
+
+  /* Ignoring the device number here should be fine.  */
+  return ade->st_ino % table_size;
 }
 
 static bool
@@ -187,7 +171,7 @@ hash_compare_active_dir_ents (void const *x, void const *y)
 {
   struct active_dir_ent const *a = x;
   struct active_dir_ent const *b = y;
-  return a->inum == b->inum;
+  return SAME_INODE (*a, *b);
 }
 
 /* A hash function for null-terminated char* strings using
@@ -253,7 +237,7 @@ pop_dir (void)
 
   /* Pop the length stack, too.  */
   assert (obstack_object_size (&len_stack) >= sizeof (size_t));
-  obstack_blank (&len_stack, -(sizeof (size_t)));
+  obstack_blank (&len_stack, (int) -(sizeof (size_t)));
 }
 
 /* Copy the SRC_LEN bytes of data beginning at SRC into the DST_LEN-byte
@@ -346,23 +330,29 @@ full_filename (const char *filename)
   return buf;
 }
 
+static inline void
+fspec_init_common (struct File_spec *fs)
+{
+  fs->have_full_mode = 0;
+  fs->have_filetype_mode = 0;
+  fs->have_device = 0;
+}
+
 void
 fspec_init_file (struct File_spec *fs, const char *filename)
 {
   fs->filename = (char *) filename;
-  fs->have_full_mode = 0;
-  fs->have_filetype_mode = 0;
+  fspec_init_common (fs);
 }
 
 static inline void
 fspec_init_dp (struct File_spec *fs, struct dirent *dp)
 {
   fs->filename = dp->d_name;
-  fs->have_full_mode = 0;
-  fs->have_filetype_mode = 0;
-  fs->inum = D_INO (dp);
+  fspec_init_common (fs);
+  fs->st_ino = D_INO (dp);
 
-#if D_TYPE_IN_DIRENT && defined (DT_UNKNOWN) && defined (DTTOIF)
+#if D_TYPE_IN_DIRENT && defined DT_UNKNOWN && defined DTTOIF
   if (dp->d_type != DT_UNKNOWN)
     {
       fs->have_filetype_mode = 1;
@@ -372,15 +362,12 @@ fspec_init_dp (struct File_spec *fs, struct dirent *dp)
 }
 
 static inline int
-fspec_get_full_mode (struct File_spec *fs, mode_t *full_mode)
+fspec_get_full_mode (struct File_spec *fs)
 {
   struct stat stat_buf;
 
   if (fs->have_full_mode)
-    {
-      *full_mode = fs->mode;
-      return 0;
-    }
+    return 0;
 
   if (lstat (fs->filename, &stat_buf))
     return 1;
@@ -388,9 +375,31 @@ fspec_get_full_mode (struct File_spec *fs, mode_t *full_mode)
   fs->have_full_mode = 1;
   fs->have_filetype_mode = 1;
   fs->mode = stat_buf.st_mode;
-  fs->inum = stat_buf.st_ino;
+  fs->st_ino = stat_buf.st_ino;
+  fs->have_device = 1;
+  fs->st_dev = stat_buf.st_dev;
+
+  return 0;
+}
+
+static inline int
+fspec_get_device_number (struct File_spec *fs)
+{
+  struct stat stat_buf;
+
+  if (fs->have_device)
+    return 0;
+
+  if (lstat (fs->filename, &stat_buf))
+    return 1;
+
+  fs->have_full_mode = 1;
+  fs->have_filetype_mode = 1;
+  fs->mode = stat_buf.st_mode;
+  fs->st_ino = stat_buf.st_ino;
+  fs->have_device = 1;
+  fs->st_dev = stat_buf.st_dev;
 
-  *full_mode = stat_buf.st_mode;
   return 0;
 }
 
@@ -399,15 +408,9 @@ fspec_get_filetype_mode (struct File_spec *fs, mode_t *filetype_mode)
 {
   int fail;
 
-  if (fs->have_filetype_mode)
-    {
-      *filetype_mode = fs->mode;
-      fail = 0;
-    }
-  else
-    {
-      fail = fspec_get_full_mode (fs, filetype_mode);
-    }
+  fail = fs->have_filetype_mode ? 0 : fspec_get_full_mode (fs);
+  if (!fail)
+    *filetype_mode = fs->mode;
 
   return fail;
 }
@@ -460,7 +463,7 @@ remove_cwd_entries (const struct rm_options *x)
       if (CLOSEDIR (dirp))
 	{
 	  /* FIXME-someday: but this is actually the previously opened dir.  */
-	  error (0, errno, "%s", full_filename ("."));
+	  error (0, errno, "%s", quote (full_filename (".")));
 	  status = RM_ERROR;
 	}
       dirp = NULL;
@@ -476,7 +479,8 @@ remove_cwd_entries (const struct rm_options *x)
 	{
 	  if (errno != ENOENT || !x->ignore_missing_files)
 	    {
-	      error (0, errno, "%s", full_filename ("."));
+	      error (0, errno, _("cannot open directory %s"),
+		     quote (full_filename (".")));
 	      status = RM_ERROR;
 	    }
 	  break;
@@ -557,18 +561,15 @@ remove_cwd_entries (const struct rm_options *x)
 	     we don't consider it again if we reopen this directory later.  */
 	  if (status != RM_OK)
 	    {
-	      bool done;
-
 	      if (ht == NULL)
 		{
-		  ht = hash_initialize (HT_INITIAL_CAPACITY, hash_pjw,
+		  ht = hash_initialize (HT_INITIAL_CAPACITY, NULL, hash_pjw,
 					hash_compare_strings, NULL);
 		  if (ht == NULL)
-		    error (1, 0, _("virtual memory exhausted"));
+		    xalloc_die ();
 		}
-	      hash_insert (ht, entry_name, &done);
-	      if (!done)
-		error (1, 0, _("virtual memory exhausted"));
+	      if (! hash_insert (ht, entry_name))
+		xalloc_die ();
 	    }
 	  else
 	    {
@@ -586,7 +587,8 @@ remove_cwd_entries (const struct rm_options *x)
     {
       if (CLOSEDIR (dirp))
 	{
-	  error (0, errno, "%s", full_filename ("."));
+	  error (0, errno, _("closing directory %s"),
+		 quote (full_filename (".")));
 	  status = RM_ERROR;
 	}
       dirp = NULL;
@@ -620,9 +622,9 @@ remove_file (struct File_spec *fs, const struct rm_options *x)
 	{
 	  fprintf (stderr,
 		   (S_ISDIR (fspec_filetype_mode (fs))
-		    ? _("%s: remove write-protected directory `%s'? ")
-		    : _("%s: remove write-protected file `%s'? ")),
-		   program_name, full_filename (pathname));
+		    ? _("%s: remove write-protected directory %s? ")
+		    : _("%s: remove write-protected file %s? ")),
+		   program_name, quote (full_filename (pathname)));
 	  if (!yesno ())
 	    return RM_USER_DECLINED;
 
@@ -637,19 +639,19 @@ remove_file (struct File_spec *fs, const struct rm_options *x)
 	 this file.  */
       fprintf (stderr,
 	       (S_ISDIR (fspec_filetype_mode (fs))
-		? _("%s: remove directory `%s'? ")
-		: _("%s: remove `%s'? ")),
-	       program_name, full_filename (pathname));
+		? _("%s: remove directory %s? ")
+		: _("%s: remove %s? ")),
+	       program_name, quote (full_filename (pathname)));
       if (!yesno ())
 	return RM_USER_DECLINED;
     }
 
   if (x->verbose)
-    printf (_("removing %s\n"), full_filename (pathname));
+    printf (_("removing %s\n"), quote (full_filename (pathname)));
 
   if (unlink (pathname) && (errno != ENOENT || !x->ignore_missing_files))
     {
-      error (0, errno, _("cannot unlink `%s'"), full_filename (pathname));
+      error (0, errno, _("cannot unlink %s"), quote (full_filename (pathname)));
       return RM_ERROR;
     }
   return RM_OK;
@@ -671,30 +673,30 @@ remove_dir (struct File_spec *fs, int need_save_cwd, const struct rm_options *x)
 
   if (!x->recursive)
     {
-      error (0, 0, _("%s: is a directory"), full_filename (dir_name));
+      error (0, 0, _("%s is a directory"), quote (full_filename (dir_name)));
       return RM_ERROR;
     }
 
   if (!x->ignore_missing_files && (x->interactive || x->stdin_tty)
       && euidaccess (dir_name, W_OK))
     {
-      fmt = _("%s: directory `%s' is write protected; descend into it anyway? ");
+      fmt = _("%s: directory %s is write protected; descend into it anyway? ");
     }
   else if (x->interactive)
     {
-      fmt = _("%s: descend into directory `%s'? ");
+      fmt = _("%s: descend into directory %s? ");
     }
 
   if (fmt)
     {
-      fprintf (stderr, fmt, program_name, full_filename (dir_name));
+      fprintf (stderr, fmt, program_name, quote (full_filename (dir_name)));
       if (!yesno ())
 	return RM_USER_DECLINED;
     }
 
   if (x->verbose)
     printf (_("removing all entries of directory %s\n"),
-	    full_filename (dir_name));
+	    quote (full_filename (dir_name)));
 
   /* Save cwd if needed.  */
   if (need_save_cwd && save_cwd (&cwd))
@@ -704,12 +706,40 @@ remove_dir (struct File_spec *fs, int need_save_cwd, const struct rm_options *x)
   if (chdir (dir_name) < 0)
     {
       error (0, errno, _("cannot change to directory %s"),
-	     full_filename (dir_name));
+	     quote (full_filename (dir_name)));
       if (need_save_cwd)
 	free_cwd (&cwd);
       return RM_ERROR;
     }
 
+  /* Verify that the device and inode numbers of `.' are the same as
+     the ones we recorded for dir_name before we cd'd into it.  This
+     detects the scenario in which an attacker tries to make Bob's rm
+     command remove some other directory belonging to Bob.  The method
+     would be to replace an existing lstat'd but-not-yet-removed directory
+     with a symlink to the target directory.  */
+  {
+    struct stat sb;
+    if (lstat (".", &sb))
+      error (EXIT_FAILURE, errno,
+	     _("cannot lstat `.' in %s"), quote (full_filename (dir_name)));
+
+    assert (fs->have_device);
+    if (!SAME_INODE (sb, *fs))
+      {
+	error (EXIT_FAILURE, 0,
+	       _("ERROR: the directory %s initially had device/inode\n\
+numbers %lu/%lu, but now (after a chdir into it), the numbers for `.'\n\
+are %lu/%lu.  That means that while rm was running, the directory\n\
+was replaced with either another directory or a link to another directory."),
+	       quote (full_filename (dir_name)),
+	       (unsigned long)(fs->st_dev),
+	       (unsigned long)(fs->st_ino),
+	       (unsigned long)(sb.st_dev),
+	       (unsigned long)(sb.st_ino));
+      }
+  }
+
   push_dir (dir_name);
 
   /* Save a copy of dir_name.  Otherwise, remove_cwd_entries may clobber
@@ -734,15 +764,15 @@ remove_dir (struct File_spec *fs, int need_save_cwd, const struct rm_options *x)
   else if (chdir ("..") < 0)
     {
       error (0, errno, _("cannot change back to directory %s via `..'"),
-	     full_filename (dir_name));
+	     quote (full_filename (dir_name)));
       return RM_ERROR;
     }
 
   if (x->interactive)
     {
-      fprintf (stderr, _("%s: remove directory `%s'%s? "),
+      fprintf (stderr, _("%s: remove directory %s%s? "),
 	       program_name,
-	       full_filename (dir_name),
+	       quote (full_filename (dir_name)),
 	       (status != RM_OK ? _(" (might be nonempty)") : ""));
       if (!yesno ())
 	{
@@ -751,7 +781,8 @@ remove_dir (struct File_spec *fs, int need_save_cwd, const struct rm_options *x)
     }
 
   if (x->verbose)
-    printf (_("removing the directory itself: %s\n"), full_filename (dir_name));
+    printf (_("removing the directory itself: %s\n"),
+	    quote (full_filename (dir_name)));
 
   if (rmdir (dir_name) && (errno != ENOENT || !x->ignore_missing_files))
     {
@@ -765,13 +796,13 @@ remove_dir (struct File_spec *fs, int need_save_cwd, const struct rm_options *x)
 	 `...': Invalid argument'  */
       if (errno == EINVAL && same_file (".", dir_name))
 	{
-	  error (0, 0, _("cannot remove current directory `%s'"),
-		 full_filename (dir_name));
+	  error (0, 0, _("cannot remove current directory %s"),
+		 quote (full_filename (dir_name)));
 	}
       else
 	{
-	  error (0, saved_errno, _("cannot remove directory `%s'"),
-		 full_filename (dir_name));
+	  error (0, saved_errno, _("cannot remove directory %s"),
+		 quote (full_filename (dir_name)));
 	}
       return RM_ERROR;
     }
@@ -792,6 +823,8 @@ rm (struct File_spec *fs, int user_specified_name, const struct rm_options *x)
 
   if (user_specified_name)
     {
+      /* CAUTION: this use of base_name works only because any
+	 trailing slashes in fs->filename have already been removed. */
       char *base = base_name (fs->filename);
 
       if (DOT_OR_DOTDOT (base))
@@ -806,30 +839,28 @@ rm (struct File_spec *fs, int user_specified_name, const struct rm_options *x)
       if (x->ignore_missing_files && errno == ENOENT)
 	return RM_OK;
 
-      error (0, errno, _("cannot remove `%s'"), full_filename (fs->filename));
+      error (0, errno, _("cannot remove %s"),
+	     quote (full_filename (fs->filename)));
       return RM_ERROR;
     }
 
 #ifdef ENABLE_CYCLE_CHECK
   if (S_ISDIR (filetype_mode))
     {
-      bool done;
       struct active_dir_ent *old_ent;
+      struct active_dir_ent *new_ent;
 
-      /* Insert this directory in the active_dir_map.
-	 If there is already a directory in the map with the same inum,
-	 then there's *probably* a directory cycle.  This test can get
-	 a false positive if two directories have the same inode number
-	 but different device numbers and one directory contains the
-	 other.  But since people don't often try to delete hierarchies
-	 containing mount points, and when they do, duplicate inode
-	 numbers are not that likely, this isn't worth detecting.  */
-      old_ent = hash_insert (active_dir_map,
-			     make_active_dir_ent (fs->inum, current_depth ()),
-			     &done);
-      if (!done)
-	  error (1, 0, _("virtual memory exhausted"));
+      /* If there is already a directory in the map with the same device
+	 and inode numbers, then there is a directory cycle.  */
 
+      if (fspec_get_device_number (fs))
+	{
+	  error (0, errno, _("cannot stat %s"),
+		 quote (full_filename (fs->filename)));
+	  return RM_ERROR;
+	}
+      new_ent = make_active_dir_ent (fs->st_ino, fs->st_dev, current_depth ());
+      old_ent = hash_lookup (active_dir_map, new_ent);
       if (old_ent)
 	{
 	  error (0, 0, _("\
@@ -837,15 +868,10 @@ WARNING: Circular directory structure.\n\
 This almost certainly means that you have a corrupted file system.\n\
 NOTIFY YOUR SYSTEM MANAGER.\n\
 The following two directories have the same inode number:\n"));
-	  /* FIXME: test this!!  */
-	  print_nth_dir (stderr, current_depth ());
-	  fputc ('\n', stderr);
 	  print_nth_dir (stderr, old_ent->depth);
-	  fputc ('\n', stderr);
+	  fprintf (stderr, "\n%s\n", quote (full_filename (fs->filename)));
 	  fflush (stderr);
 
-	  free (old_ent);
-
 	  if (x->interactive)
 	    {
 	      error (0, 0, _("continue? "));
@@ -854,6 +880,10 @@ The following two directories have the same inode number:\n"));
 	    }
 	  exit (1);
 	}
+
+      /* Put this directory in the active_dir_map.  */
+      if (! hash_insert (active_dir_map, new_ent))
+	xalloc_die ();
     }
 #endif
 
@@ -877,7 +907,9 @@ The following two directories have the same inode number:\n"));
 	struct active_dir_ent *old_ent;
 
 	/* Remove this directory from the active_dir_map.  */
-	tmp.inum = fs->inum;
+	tmp.st_ino = fs->st_ino;
+	assert (fs->have_device);
+	tmp.st_dev = fs->st_dev;
 	old_ent = hash_delete (active_dir_map, &tmp);
 	assert (old_ent != NULL);
 	free (old_ent);
@@ -896,7 +928,7 @@ remove_init (void)
   obstack_init (&len_stack);
 
 #ifdef ENABLE_CYCLE_CHECK
-  active_dir_map = hash_initialize (ACTIVE_DIR_INITIAL_CAPACITY,
+  active_dir_map = hash_initialize (ACTIVE_DIR_INITIAL_CAPACITY, NULL,
 				    hash_active_dir_ent,
 				    hash_compare_active_dir_ents, free);
 #endif
