@@ -1,5 +1,5 @@
 /* sync - update the super block
-   Copyright (C) 94, 95, 96, 97, 1998 Free Software Foundation, Inc.
+   Copyright (C) 1994-2000 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -15,21 +15,25 @@
    along with this program; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
 
-/* Jim Meyering <meyering@comco.com> */
+/* Written by Jim Meyering */
 
 #include <config.h>
 #include <stdio.h>
 #include <sys/types.h>
 
 #include "system.h"
-#include "long-options.h"
-#include "closeout.h"
 #include "error.h"
+#include "long-options.h"
+
+/* The official name of this program (e.g., no `g' prefix).  */
+#define PROGRAM_NAME "sync"
+
+#define AUTHORS "Jim Meyering"
 
 /* The name this program was run with. */
 char *program_name;
 
-static void
+void
 usage (int status)
 {
   if (status != 0)
@@ -45,7 +49,6 @@ Force changed blocks to disk, update the super block.\n\
   --version   output version information and exit\n\
 "));
       puts (_("\nReport bugs to <bug-fileutils@gnu.org>."));
-      close_stdout ();
     }
   exit (status);
 }
@@ -58,7 +61,10 @@ main (int argc, char **argv)
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
-  parse_long_options (argc, argv, "sync", GNU_PACKAGE, VERSION, usage);
+  atexit (close_stdout);
+
+  parse_long_options (argc, argv, PROGRAM_NAME, GNU_PACKAGE, VERSION,
+		      AUTHORS, usage);
 
   if (argc != 1)
     error (0, 0, _("ignoring all arguments"));
