@@ -1,5 +1,5 @@
 /* touch -- change modification and access times of files
-   Copyright (C) 87, 89, 90, 91, 95, 96, 97, 1998 Free Software Foundation, Inc.
+   Copyright (C) 87, 1989-1991, 1995-2000 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -24,24 +24,39 @@
 #include <sys/types.h>
 
 #include "system.h"
-#include "closeout.h"
-#include "error.h"
 #include "argmatch.h"
+#include "error.h"
 #include "getdate.h"
-#include "safe-read.h"
 #include "posixtm.h"
+#include "quote.h"
+#include "safe-read.h"
+
+/* The official name of this program (e.g., no `g' prefix).  */
+#define PROGRAM_NAME "touch"
+
+#define AUTHORS \
+  "Paul Rubin, Arnold Robbins, Jim Kingdon, David MacKenzie, and Randy Smith"
 
 #ifndef STDC_HEADERS
 time_t time ();
 #endif
 
-int full_write ();
-void invalid_arg ();
-
 /* Bitmasks for `change_times'. */
 #define CH_ATIME 1
 #define CH_MTIME 2
 
+#if !defined O_NDELAY
+# define O_NDELAY 0
+#endif
+
+#if !defined O_NONBLOCK
+# define O_NONBLOCK O_NDELAY
+#endif
+
+#if !defined O_NOCTTY
+# define O_NOCTTY 0
+#endif
+
 /* The name by which this program was run. */
 char *program_name;
 
@@ -62,7 +77,9 @@ static int posix_date;
 
 /* If nonzero, the only thing we have to do is change both the
    modification and access time to the current time, so we don't
-   have to own the file, just be able to read and write it.  */
+   have to own the file, just be able to read and write it.
+   On some systems, we can do this if we own the file, even though
+   we have neither read nor write access to it.  */
 static int amtime_now;
 
 /* New time to use when setting time. */
@@ -74,21 +91,22 @@ static char *ref_file;
 /* Info about the reference file. */
 static struct stat ref_stats;
 
-/* If nonzero, display usage information and exit.  */
-static int show_help;
-
-/* If nonzero, print the version on standard output and exit.  */
-static int show_version;
+/* For long options that have no equivalent short option, use a
+   non-character as a pseudo short option, starting with CHAR_MAX + 1.  */
+enum
+{
+  TIME_OPTION = CHAR_MAX + 1
+};
 
 static struct option const longopts[] =
 {
-  {"time", required_argument, 0, CHAR_MAX + 1},
+  {"time", required_argument, 0, TIME_OPTION},
   {"no-create", no_argument, 0, 'c'},
   {"date", required_argument, 0, 'd'},
-  {"file", required_argument, 0, 'r'},
+  {"file", required_argument, 0, 'r'}, /* FIXME: phase out --file */
   {"reference", required_argument, 0, 'r'},
-  {"help", no_argument, &show_help, 1},
-  {"version", no_argument, &show_version, 1},
+  {GETOPT_HELP_OPTION_DECL},
+  {GETOPT_VERSION_OPTION_DECL},
   {0, 0, 0, 0}
 };
 
@@ -108,47 +126,43 @@ static int const time_masks[] =
    Return 0 if successful, 1 if an error occurs. */
 
 static int
-touch (char *file)
+touch (const char *file)
 {
   int status;
   struct stat sbuf;
-  int fd;
+  int fd = -1;
+  int open_errno = 0;
 
-  if (stat (file, &sbuf))
+  if (! no_create)
     {
-      if (errno != ENOENT)
-	{
-	  error (0, errno, "%s", file);
-	  return 1;
-	}
-      if (no_create)
-	return 0;
-      fd = creat (file, 0666);
+      /* Try to open FILE, creating it if necessary.  */
+      fd = open (file, O_WRONLY | O_CREAT | O_NONBLOCK | O_NOCTTY,
+		 S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
       if (fd == -1)
+	open_errno = errno;
+    }
+
+  if (! amtime_now)
+    {
+      /* We're setting only one of the time values.  stat the target to get
+	 the other one.  If we have the file descriptor already, use fstat.
+	 Otherwise, either we're in no-create mode (and hence didn't call open)
+	 or FILE is inaccessible or a directory, so we have to use stat.  */
+      if (fd != -1 ? fstat (fd, &sbuf) : stat (file, &sbuf))
 	{
-	  error (0, errno, "%s", file);
-	  return 1;
-	}
-      if (amtime_now)
-	{
-	  if (close (fd) < 0)
-	    {
-	      error (0, errno, "%s", file);
-	      return 1;
-	    }
-	  return 0;		/* We've done all we have to. */
-	}
-      if (fstat (fd, &sbuf))
-	{
-	  error (0, errno, "%s", file);
+	  if (open_errno)
+	    error (0, open_errno, _("creating %s"), quote (file));
+	  else
+	    error (0, errno, _("getting attributes of %s"), quote (file));
 	  close (fd);
 	  return 1;
 	}
-      if (close (fd) < 0)
-	{
-	  error (0, errno, "%s", file);
-	  return 1;
-	}
+    }
+
+  if (fd != -1 && close (fd) < 0)
+    {
+      error (0, errno, _("creating %s"), quote (file));
+      return 1;
     }
 
   if (amtime_now)
@@ -184,14 +198,17 @@ touch (char *file)
 
   if (status)
     {
-      error (0, errno, "%s", file);
+      if (open_errno)
+	error (0, open_errno, _("creating %s"), quote (file));
+      else
+	error (0, errno, _("setting times of %s"), quote (file));
       return 1;
     }
 
   return 0;
 }
 
-static void
+void
 usage (int status)
 {
   if (status != 0)
@@ -206,22 +223,21 @@ usage (int status)
 Update the access and modification times of each FILE to the current time.\n\
 \n\
   -a                     change only the access time\n\
-  -c                     do not create any files\n\
+  -c, --no-create        do not create any files\n\
   -d, --date=STRING      parse STRING and use it instead of current time\n\
   -f                     (ignored)\n\
   -m                     change only the modification time\n\
   -r, --reference=FILE   use this file's times instead of current time\n\
   -t STAMP               use [[CC]YY]MMDDhhmm[.ss] instead of current time\n\
-      --time=WORD        access -a, atime -a, mtime -m, modify -m, use -a\n\
+  --time=WORD            set time given by WORD: access atime use (same as -a)\n\
+                           modify mtime (same as -m)\n\
       --help             display this help and exit\n\
       --version          output version information and exit\n\
 \n\
-STAMP may be used without -t if none of -drt, nor --, are used.\n\
 Note that the three time-date formats recognized for the -d and -t options\n\
 and for the obsolescent argument are all different.\n\
 "));
       puts (_("\nReport bugs to <bug-fileutils@gnu.org>."));
-      close_stdout ();
     }
   exit (status);
 }
@@ -229,7 +245,7 @@ and for the obsolescent argument are all different.\n\
 int
 main (int argc, char **argv)
 {
-  int c, i;
+  int c;
   int date_set = 0;
   int err = 0;
 
@@ -238,6 +254,8 @@ main (int argc, char **argv)
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
+  atexit (close_stdout);
+
   change_times = no_create = use_ref = posix_date = flexible_date = 0;
   newtime = (time_t) -1;
 
@@ -260,7 +278,7 @@ main (int argc, char **argv)
 	  flexible_date++;
 	  newtime = get_date (optarg, NULL);
 	  if (newtime == (time_t) -1)
-	    error (1, 0, _("invalid date format `%s'"), optarg);
+	    error (1, 0, _("invalid date format %s"), quote (optarg));
 	  date_set++;
 	  break;
 
@@ -281,35 +299,24 @@ main (int argc, char **argv)
 	  newtime = posixtime (optarg,
 			       PDS_LEADING_YEAR | PDS_CENTURY | PDS_SECONDS);
 	  if (newtime == (time_t) -1)
-	    error (1, 0, _("invalid date format `%s'"), optarg);
+	    error (1, 0, _("invalid date format %s"), quote (optarg));
 	  date_set++;
 	  break;
 
-	case CHAR_MAX + 1:
-	  i = argmatch (optarg, time_args);
-	  if (i < 0)
-	    {
-	      invalid_arg (_("time selector"), optarg, i);
-	      usage (1);
-	    }
-	  change_times |= time_masks[i];
+	case TIME_OPTION:	/* --time */
+	  change_times |= XARGMATCH ("--time", optarg,
+				     time_args, time_masks);
 	  break;
 
+	case_GETOPT_HELP_CHAR;
+
+	case_GETOPT_VERSION_CHAR (PROGRAM_NAME, AUTHORS);
+
 	default:
 	  usage (1);
 	}
     }
 
-  if (show_version)
-    {
-      printf ("touch (%s) %s\n", GNU_PACKAGE, VERSION);
-      close_stdout ();
-      exit (0);
-    }
-
-  if (show_help)
-    usage (0);
-
   if (change_times == 0)
     change_times = CH_ATIME | CH_MTIME;
 
@@ -323,7 +330,7 @@ main (int argc, char **argv)
   if (use_ref)
     {
       if (stat (ref_file, &ref_stats))
-	error (1, errno, "%s", ref_file);
+	error (1, errno, _("getting attributes of %s"), quote (ref_file));
       date_set++;
     }
 
